"""Drop-in replacement for Python's difflib module, backed by Go's diff package.

Uses Go FFI for unified diff generation on large inputs.
Falls back to Python for SequenceMatcher, get_close_matches, etc.

Usage:
    from goated.compat import difflib
    diff = difflib.unified_diff(a_lines, b_lines, fromfile='a', tofile='b')
"""

from __future__ import annotations

import ctypes
import difflib as _difflib

# Re-export everything from Python's difflib
from difflib import (  # noqa: F401
    Differ,
    HtmlDiff,
    SequenceMatcher,
    context_diff,
    get_close_matches,
    ndiff,
    restore,
)
from typing import Sequence

from goated._core import _USE_GO_LIB, get_lib

_GO_THRESHOLD = 50  # Use Go for diffs with > 50 lines
_lib_setup = False


def _setup_lib() -> None:
    global _lib_setup
    if _lib_setup or not _USE_GO_LIB:
        return
    try:
        lib = get_lib().lib
        lib.goated_diff_unified.argtypes = [
            ctypes.POINTER(ctypes.c_char_p),  # aLines
            ctypes.c_int,                      # aCount
            ctypes.POINTER(ctypes.c_char_p),  # bLines
            ctypes.c_int,                      # bCount
            ctypes.c_char_p,                   # fromFile
            ctypes.c_char_p,                   # toFile
            ctypes.c_int,                      # contextLines
            ctypes.POINTER(ctypes.c_int),      # outLen
        ]
        lib.goated_diff_unified.restype = ctypes.c_char_p
        _lib_setup = True
    except Exception:
        pass


def unified_diff(
    a: Sequence[str],
    b: Sequence[str],
    fromfile: str = "",
    tofile: str = "",
    fromfiledate: str = "",
    tofiledate: str = "",
    n: int = 3,
    lineterm: str = "\n",
) -> list[str]:
    """Generate unified diff. Go-accelerated for large inputs.

    Unlike stdlib which returns a generator, this returns a list for
    Go batch processing. The result is compatible with list(difflib.unified_diff(...)).
    """
    a_list = list(a)
    b_list = list(b)

    if len(a_list) + len(b_list) < _GO_THRESHOLD or not _USE_GO_LIB:
        return list(_difflib.unified_diff(
            a_list, b_list, fromfile=fromfile, tofile=tofile,
            fromfiledate=fromfiledate, tofiledate=tofiledate,
            n=n, lineterm=lineterm,
        ))

    # filedate not supported in Go path; fall back if used
    if fromfiledate or tofiledate:
        return list(_difflib.unified_diff(
            a_list, b_list, fromfile=fromfile, tofile=tofile,
            fromfiledate=fromfiledate, tofiledate=tofiledate,
            n=n, lineterm=lineterm,
        ))

    # Prefer cffi
    from goated._core import get_cffi_lib

    cffi_lib = get_cffi_lib()
    if cffi_lib is not None:
        try:
            from goated._core import _cffi_ffi

            a_encoded = [line.encode("utf-8") for line in a_list]
            b_encoded = [line.encode("utf-8") for line in b_list]
            a_arr = _cffi_ffi.new("char*[]", [_cffi_ffi.new("char[]", s) for s in a_encoded])
            b_arr = _cffi_ffi.new("char*[]", [_cffi_ffi.new("char[]", s) for s in b_encoded])
            out_len = _cffi_ffi.new("int*")

            result = cffi_lib.goated_diff_unified(
                a_arr, len(a_list), b_arr, len(b_list),
                fromfile.encode("utf-8"), tofile.encode("utf-8"),
                n, out_len,
            )
            if result:
                text = _cffi_ffi.string(result).decode("utf-8")
                if text:
                    return text.split("\n")
                return []
        except Exception:
            pass

    # Fallback to ctypes
    _setup_lib()
    if _lib_setup:
        try:
            lib = get_lib().lib
            a_encoded = [line.encode("utf-8") for line in a_list]
            b_encoded = [line.encode("utf-8") for line in b_list]
            a_arr = (ctypes.c_char_p * len(a_list))(*a_encoded)
            b_arr = (ctypes.c_char_p * len(b_list))(*b_encoded)
            out_len = ctypes.c_int()

            result = lib.goated_diff_unified(
                a_arr, len(a_list), b_arr, len(b_list),
                fromfile.encode("utf-8"), tofile.encode("utf-8"),
                n, ctypes.byref(out_len),
            )
            if result:
                text = result.decode("utf-8")
                if text:
                    return text.split("\n")
                return []
        except Exception:
            pass

    return list(_difflib.unified_diff(
        a_list, b_list, fromfile=fromfile, tofile=tofile,
        n=n, lineterm=lineterm,
    ))


__all__ = [
    "unified_diff",
    "context_diff",
    "ndiff",
    "restore",
    "Differ",
    "HtmlDiff",
    "SequenceMatcher",
    "get_close_matches",
]



from __future__ import annotations

import struct
from typing import IO, Any

from goated.result import Err, GoError, Ok, Result

__all__ = [
    "Read",
    "Write",
    "Size",
    "PutVarint",
    "PutUvarint",
    "Varint",
    "Uvarint",
    "ReadVarint",
    "ReadUvarint",
    "BigEndian",
    "LittleEndian",
    "ByteOrder",
    "MaxVarintLen16",
    "MaxVarintLen32",
    "MaxVarintLen64",
]

MaxVarintLen16 = 3
MaxVarintLen32 = 5
MaxVarintLen64 = 10


class ByteOrder:
    """ByteOrder specifies how to convert byte slices into integers."""

    def __init__(self, name: str, format_prefix: str):
        self._name = name
        self._prefix = format_prefix

    def Uint16(self, b: bytes) -> int:
        """Uint16 returns the uint16 in b."""
        result: int = struct.unpack(f"{self._prefix}H", b[:2])[0]
        return result

    def Uint32(self, b: bytes) -> int:
        """Uint32 returns the uint32 in b."""
        result: int = struct.unpack(f"{self._prefix}I", b[:4])[0]
        return result

    def Uint64(self, b: bytes) -> int:
        """Uint64 returns the uint64 in b."""
        result: int = struct.unpack(f"{self._prefix}Q", b[:8])[0]
        return result

    def PutUint16(self, b: bytearray, v: int) -> None:
        """PutUint16 encodes v into b."""
        data = struct.pack(f"{self._prefix}H", v)
        b[:2] = data

    def PutUint32(self, b: bytearray, v: int) -> None:
        """PutUint32 encodes v into b."""
        data = struct.pack(f"{self._prefix}I", v)
        b[:4] = data

    def PutUint64(self, b: bytearray, v: int) -> None:
        """PutUint64 encodes v into b."""
        data = struct.pack(f"{self._prefix}Q", v)
        b[:8] = data

    def AppendUint16(self, b: bytes, v: int) -> bytes:
        """AppendUint16 appends v to b."""
        return b + struct.pack(f"{self._prefix}H", v)

    def AppendUint32(self, b: bytes, v: int) -> bytes:
        """AppendUint32 appends v to b."""
        return b + struct.pack(f"{self._prefix}I", v)

    def AppendUint64(self, b: bytes, v: int) -> bytes:
        """AppendUint64 appends v to b."""
        return b + struct.pack(f"{self._prefix}Q", v)

    def String(self) -> str:
        """String returns the name of the byte order."""
        return self._name


BigEndian = ByteOrder("BigEndian", ">")
LittleEndian = ByteOrder("LittleEndian", "<")


def Read(r: IO[bytes], order: ByteOrder, data: Any) -> Result[None, GoError]:
    """Read reads structured binary data from r into data."""
    try:
        if isinstance(data, list):
            for i, item in enumerate(data):
                if isinstance(item, int):
                    b = r.read(8)
                    if len(b) < 8:
                        return Err(GoError("unexpected EOF", "io.EOF"))
                    data[i] = order.Uint64(b)
        elif hasattr(data, "__dict__"):
            for name, value in data.__dict__.items():
                if isinstance(value, int):
                    b = r.read(8)
                    if len(b) < 8:
                        return Err(GoError("unexpected EOF", "io.EOF"))
                    setattr(data, name, order.Uint64(b))
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "binary.ReadError"))


def Write(w: IO[bytes], order: ByteOrder, data: Any) -> Result[None, GoError]:
    """Write writes the binary representation of data to w."""
    try:
        if isinstance(data, bytes):
            w.write(data)
        elif isinstance(data, int):
            if data < 0:
                w.write(struct.pack(f"{order._prefix}q", data))
            elif data <= 0xFF:
                w.write(struct.pack("B", data))
            elif data <= 0xFFFF:
                w.write(struct.pack(f"{order._prefix}H", data))
            elif data <= 0xFFFFFFFF:
                w.write(struct.pack(f"{order._prefix}I", data))
            else:
                w.write(struct.pack(f"{order._prefix}Q", data))
        elif isinstance(data, float):
            w.write(struct.pack(f"{order._prefix}d", data))
        elif isinstance(data, (list, tuple)):
            for item in data:
                Write(w, order, item)
        elif hasattr(data, "__dict__"):
            for value in data.__dict__.values():
                Write(w, order, value)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "binary.WriteError"))


def Size(v: Any) -> int:
    """Size returns how many bytes Write would generate to encode the value v."""
    if isinstance(v, bytes):
        return len(v)
    elif isinstance(v, int):
        if v < 0 or v > 0xFFFFFFFF:
            return 8
        elif v <= 0xFF:
            return 1
        elif v <= 0xFFFF:
            return 2
        else:
            return 4
    elif isinstance(v, float):
        return 8
    elif isinstance(v, (list, tuple)):
        return sum(Size(item) for item in v)
    elif hasattr(v, "__dict__"):
        return sum(Size(val) for val in v.__dict__.values())
    return 0


def PutVarint(buf: bytearray, x: int) -> int:
    """PutVarint encodes an int64 into buf and returns the number of bytes written."""
    ux = (x << 1) ^ (x >> 63)
    return PutUvarint(buf, ux)


def PutUvarint(buf: bytearray, x: int) -> int:
    """PutUvarint encodes a uint64 into buf and returns the number of bytes written."""
    i = 0
    while x >= 0x80:
        buf[i] = (x & 0x7F) | 0x80
        x >>= 7
        i += 1
    buf[i] = x & 0xFF
    return i + 1


def Varint(buf: bytes) -> tuple[int, int]:
    """Varint decodes an int64 from buf and returns it and the number of bytes read."""
    ux, n = Uvarint(buf)
    x = ux >> 1
    if ux & 1:
        x = ~x
    return x, n


def Uvarint(buf: bytes) -> tuple[int, int]:
    """Uvarint decodes a uint64 from buf and returns it and the number of bytes read."""
    x = 0
    s = 0
    for i, b in enumerate(buf):
        if i == MaxVarintLen64:
            return 0, -(i + 1)
        if b < 0x80:
            if i == MaxVarintLen64 - 1 and b > 1:
                return 0, -(i + 1)
            return x | (b << s), i + 1
        x |= (b & 0x7F) << s
        s += 7
    return 0, 0


def ReadVarint(r: IO[bytes]) -> Result[int, GoError]:
    """ReadVarint reads an encoded signed integer from r."""
    result = ReadUvarint(r)
    if result.is_err():
        return result
    ux = result.unwrap()
    x = ux >> 1
    if ux & 1:
        x = ~x
    return Ok(x)


def ReadUvarint(r: IO[bytes]) -> Result[int, GoError]:
    """ReadUvarint reads an encoded unsigned integer from r."""
    x = 0
    s = 0
    for i in range(MaxVarintLen64):
        b = r.read(1)
        if not b:
            return Err(GoError("unexpected EOF", "io.EOF"))
        byte = b[0]
        if byte < 0x80:
            if i == MaxVarintLen64 - 1 and byte > 1:
                return Err(GoError("overflow", "binary.Overflow"))
            return Ok(x | (byte << s))
        x |= (byte & 0x7F) << s
        s += 7
    return Err(GoError("overflow", "binary.Overflow"))


def AppendVarint(b: bytes, x: int) -> bytes:
    """AppendVarint appends the varint-encoded form of x to b."""
    ux = (x << 1) ^ (x >> 63)
    return AppendUvarint(b, ux)


def AppendUvarint(b: bytes, x: int) -> bytes:
    """AppendUvarint appends the varint-encoded form of x to b."""
    result = bytearray(b)
    while x >= 0x80:
        result.append((x & 0x7F) | 0x80)
        x >>= 7
    result.append(x)
    return bytes(result)



"""Go errors package bindings - Pure Python implementation.

This module provides Python bindings for Go's errors package.

Example:
    >>> from goated.std import errors
    >>>
    >>> err = errors.New("something went wrong")
    >>> print(err)
    something went wrong
    >>>
    >>> wrapped = errors.New("operation failed: %w", err)
    >>> errors.Is(wrapped, err)
    True

"""

from __future__ import annotations

import traceback
from typing import TypeVar, cast

__all__ = [
    "New",
    "Is",
    "As",
    "Unwrap",
    "Join",
    # Error interface
    "Error",
]

T = TypeVar("T", bound=Exception)


# =============================================================================
# Error Interface
# =============================================================================


class Error(Exception):
    """Error is the interface that wraps the basic Error method.

    This is the base class for Go-style errors.
    """

    def __init__(self, message: str):
        super().__init__(message)
        self._message = message

    def Error(self) -> str:
        """Return the error message."""
        return self._message

    def __str__(self) -> str:
        return self._message

    def __repr__(self) -> str:
        return f"Error({self._message!r})"


class _WrappedError(Error):
    """An error that wraps another error."""

    def __init__(self, message: str, wrapped: Exception | None = None):
        super().__init__(message)
        self._wrapped = wrapped

    def Unwrap(self) -> Exception | None:
        """Return the wrapped error."""
        return self._wrapped


class _JoinedError(Error):
    """An error that joins multiple errors."""

    def __init__(self, errs: list[Exception]):
        self._errors = [e for e in errs if e is not None]
        messages = [str(e) for e in self._errors]
        super().__init__("\n".join(messages))

    def Unwrap(self) -> list[Exception]:
        """Return the list of wrapped errors."""
        return self._errors


# =============================================================================
# Functions
# =============================================================================


def New(text: str) -> Error:
    """New returns an error that formats as the given text.

    Each call to New returns a distinct error value even if the text is identical.

    Example:
        >>> err1 = New("error")
        >>> err2 = New("error")
        >>> err1 == err2
        False

    """
    return Error(text)


def Is(err: Exception | None, target: Exception | None) -> bool:
    """Is reports whether any error in err's tree matches target.

    The tree consists of err itself, followed by the errors obtained by
    repeatedly calling Unwrap. An error is considered to match a target
    if it is equal to that target or if it implements an Is method
    such that Is(target) returns true.

    Example:
        >>> base = New("base error")
        >>> wrapped = _WrappedError("wrapped: base error", base)
        >>> Is(wrapped, base)
        True

    """
    if err is None or target is None:
        return err == target

    # Direct comparison
    if err == target:
        return True

    # Check if error has Is method
    if hasattr(err, "Is") and callable(err.Is) and err.Is(target):
        return True

    # Check type match
    if type(err) is type(target) and str(err) == str(target):
        return True

    # Unwrap and recurse
    unwrapped = Unwrap(err)
    if unwrapped is not None:
        if isinstance(unwrapped, list):
            return any(Is(e, target) for e in unwrapped)
        return Is(unwrapped, target)

    return False


def As(err: Exception | None, target: type[T]) -> T | None:
    """As finds the first error in err's tree that matches target type,
    and if so, returns that error.

    The tree consists of err itself, followed by the errors obtained by
    repeatedly calling Unwrap.

    Example:
        >>> class MyError(Error):
        ...     pass
        >>> err = MyError("my error")
        >>> As(err, MyError)
        MyError('my error')

    """
    if err is None:
        return None

    # Check if err is of target type
    if isinstance(err, target):
        return err

    # Check if error has As method
    if hasattr(err, "As") and callable(err.As):
        result = err.As(target)
        if result is not None:
            return cast(T, result)

    # Unwrap and recurse
    unwrapped = Unwrap(err)
    if unwrapped is not None:
        if isinstance(unwrapped, list):
            for e in unwrapped:
                result = As(e, target)
                if result is not None:
                    return cast(T, result)
        else:
            return As(unwrapped, target)

    return None


def Unwrap(err: Exception | None) -> Exception | list[Exception] | None:
    """Unwrap returns the result of calling the Unwrap method on err,
    if err's type contains an Unwrap method returning error.
    Otherwise, Unwrap returns nil.

    Example:
        >>> base = New("base")
        >>> wrapped = _WrappedError("wrapped", base)
        >>> Unwrap(wrapped) == base
        True

    """
    if err is None:
        return None

    if hasattr(err, "Unwrap") and callable(err.Unwrap):
        result: Exception | list[Exception] | None = err.Unwrap()
        return result

    # Check for Python's __cause__ (from `raise X from Y`)
    if hasattr(err, "__cause__") and err.__cause__ is not None:
        cause = err.__cause__
        if isinstance(cause, Exception):
            return cause

    return None


def Join(*errs: Exception | None) -> Exception | None:
    """Join returns an error that wraps the given errors.
    Any nil error values are discarded.
    Join returns nil if errs contains no non-nil values.

    The error formats as the concatenation of the strings obtained
    by calling the Error method of each element of errs, with a newline
    between each string.

    Example:
        >>> err1 = New("error 1")
        >>> err2 = New("error 2")
        >>> joined = Join(err1, err2)
        >>> print(joined)
        error 1
        error 2

    """
    non_nil = [e for e in errs if e is not None]

    if not non_nil:
        return None

    if len(non_nil) == 1:
        return non_nil[0]

    return _JoinedError(non_nil)


# =============================================================================
# Helper Functions
# =============================================================================


def Wrap(err: Exception, message: str) -> Error:
    """Wrap returns an error annotating err with a message.

    This is a convenience function similar to fmt.Errorf("%s: %w", message, err).

    Example:
        >>> base = New("file not found")
        >>> wrapped = Wrap(base, "failed to open config")
        >>> print(wrapped)
        failed to open config: file not found

    """
    return _WrappedError(f"{message}: {err}", err)


def Cause(err: Exception | None) -> Exception | None:
    """Cause returns the underlying cause of the error, if possible.

    This unwraps the error chain to find the root cause.
    """
    if err is None:
        return None

    for _ in range(100):  # Prevent infinite loops
        unwrapped = Unwrap(err)
        if unwrapped is None:
            return err
        if isinstance(unwrapped, list):
            return err
        err = unwrapped

    return err


def WithStack(err: Exception) -> Exception:
    """WithStack annotates err with a stack trace."""
    setattr(err, "_stack", traceback.format_stack())  # noqa: B010
    return err


def Stack(err: Exception) -> str | None:
    """Stack returns the stack trace from an error created with WithStack."""
    stack = getattr(err, "_stack", None)
    if stack is not None:
        return "".join(stack)
    return None



"""Go-style concurrency API using the M:N runtime - Optimized.

This module provides the same API as goated.std.goroutine but backed
by the work-stealing M:N scheduler for better performance, especially
on free-threaded Python.

Key optimizations over standard implementation:
- Future callbacks instead of wrapper functions (zero allocation hot path)
- Atomic-like operations for WaitGroup on free-threaded Python
- Minimized function creation overhead
- Direct executor access in hot paths
- Proper condition-variable-based Select (no spin-wait)
- True unbuffered channel rendezvous semantics
- Correct reader-writer lock with writer preference
"""

from __future__ import annotations

import functools
import random
import threading
import time
from collections import deque
from collections.abc import Callable, Iterator
from concurrent.futures import Future
from concurrent.futures import wait as futures_wait
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from goated.runtime.scheduler import get_runtime

if TYPE_CHECKING:
    from goated.runtime.scheduler import Runtime

__all__ = [
    "go",
    "WaitGroup",
    "Chan",
    "Select",
    "SelectCase",
    "Mutex",
    "RWMutex",
    "Once",
    "Pool",
    "GoGroup",
    "ErrGroup",
    "goroutine",
    "Await",
    "AfterFunc",
    "Ticker",
    "After",
    "Semaphore",
    "pipe",
    "merge",
    "fan_out",
    "parallel_for",
    "parallel_map",
]

T = TypeVar("T")

# Cache for hot path - initialized on first access
_RUNTIME_CACHE: Runtime | None = None


def _get_runtime_fast() -> Runtime:
    global _RUNTIME_CACHE
    if _RUNTIME_CACHE is None:
        _RUNTIME_CACHE = get_runtime()
    return _RUNTIME_CACHE


class WaitGroup:
    """A WaitGroup waits for a collection of goroutines to finish.

    Optimized implementation using atomic-like operations where possible.

    Example:
        wg = WaitGroup()

        for i in range(10):
            wg.Add(1)
            go(worker, i, done=wg)

        wg.Wait()  # Blocks until all workers done

    """

    __slots__ = ("_count", "_lock", "_event")

    def __init__(self) -> None:
        self._count = 0
        self._lock = threading.Lock()
        self._event = threading.Event()
        self._event.set()

    def Add(self, delta: int = 1) -> None:
        """Add delta to the WaitGroup counter."""
        with self._lock:
            self._count += delta
            if self._count < 0:
                raise ValueError("negative WaitGroup counter")
            if self._count == 0:
                self._event.set()
            else:
                self._event.clear()

    def Done(self) -> None:
        """Decrement the WaitGroup counter by one."""
        # Inlined Add(-1) to avoid method call overhead
        lock = self._lock
        with lock:
            self._count -= 1
            if self._count < 0:
                raise ValueError("negative WaitGroup counter")
            if self._count == 0:
                self._event.set()

    def Wait(self, timeout: float | None = None) -> bool:
        """Block until the WaitGroup counter is zero.

        Args:
            timeout: Optional timeout in seconds

        Returns:
            True if counter reached zero, False if timeout

        """
        return self._event.wait(timeout)

    def _done_callback(self, future: Future[Any]) -> None:
        """Callback for Future completion - inlined for performance."""
        lock = self._lock
        with lock:
            self._count -= 1
            if self._count == 0:
                self._event.set()


def go(
    fn: Callable[..., Any], *args: Any, done: WaitGroup | None = None, **kwargs: Any
) -> Future[Any]:
    """Spawn a goroutine using the M:N runtime.

    Optimized to use Future callbacks instead of wrapper functions when possible.

    Args:
        fn: Function to execute
        *args: Arguments to pass to fn
        done: Optional WaitGroup to signal when done
        **kwargs: Keyword arguments to pass to fn

    Returns:
        Future that can be used to get the result

    Example:
        def work(x):
            return x * 2

        # Fire and forget
        go(work, 42)

        # With WaitGroup
        wg = WaitGroup()
        wg.Add(1)
        go(work, 42, done=wg)
        wg.Wait()

        # Get result
        future = go(work, 42)
        result = future.result()

    """
    runtime = _get_runtime_fast()

    # Fast path: no WaitGroup, no args - direct submission
    # Note: _executor is guaranteed non-None after get_runtime().start()
    executor = runtime._executor
    assert executor is not None
    if done is None:
        if not args and not kwargs:
            return executor.submit(fn)
        elif not kwargs:
            return executor.submit(fn, *args)
        else:
            return executor.submit(fn, *args, **kwargs)

    # WaitGroup path: use callback instead of wrapper (much faster!)
    if not args and not kwargs:
        future = executor.submit(fn)
    elif not kwargs:
        future = executor.submit(fn, *args)
    else:
        future = executor.submit(fn, *args, **kwargs)

    # Add callback to decrement WaitGroup when done - no wrapper needed!
    future.add_done_callback(done._done_callback)
    return future


class _ChannelClosed:
    """Sentinel for channel close signal."""

    __slots__ = ()


_CHAN_CLOSED: _ChannelClosed = _ChannelClosed()

# Global registry for Select notification: channels notify waiting Selects
_select_waiters_lock = threading.Lock()
_select_waiters: list[threading.Condition] = []


def _notify_select_waiters() -> None:
    """Wake any Select statements waiting on channel activity."""
    with _select_waiters_lock:
        for cond in _select_waiters:
            with cond:
                cond.notify_all()


class Chan(Generic[T]):
    """A typed channel for goroutine communication.

    Supports both buffered and true unbuffered (rendezvous) channels.
    Unbuffered channels (buffer=0) require sender and receiver to meet:
    the sender blocks until a receiver is ready, and vice versa.

    Optimized: uses deque + Condition (3x faster than queue.Queue).
    Iterators block properly on condition variables (no polling).

    Example:
        ch = Chan[int](buffer=5)

        # In producer goroutine
        ch.Send(42)
        ch.Close()

        # In consumer goroutine
        val, ok = ch.Recv()
        if ok:
            print(val)

        # Or iterate
        for val in ch:
            print(val)

    """

    __slots__ = (
        "_buffer",
        "_buffer_size",
        "_unbuffered",
        "_closed",
        "_lock",
        "_not_empty",
        "_not_full",
        "_rendezvous",
        "_rendezvous_value",
        "_rendezvous_taken",
    )

    def __init__(self, buffer: int = 0) -> None:
        """Create a channel.

        Args:
            buffer: Buffer size. 0 = unbuffered (synchronous rendezvous)

        """
        self._unbuffered = buffer <= 0
        self._buffer_size = max(buffer, 0)
        self._buffer: deque[T] = deque()
        self._closed = False
        self._lock = threading.Lock()
        self._not_empty = threading.Condition(self._lock)
        self._not_full = threading.Condition(self._lock)
        # Rendezvous state for unbuffered channels
        self._rendezvous: threading.Condition = threading.Condition(self._lock)
        self._rendezvous_value: T | _ChannelClosed = _CHAN_CLOSED  # sentinel = no value
        self._rendezvous_taken = False

    def Send(self, value: T, timeout: float | None = None) -> bool:
        """Send a value to the channel.

        For unbuffered channels, blocks until a receiver takes the value.
        For buffered channels, blocks only when the buffer is full.
        """
        monotonic = time.monotonic
        deadline = monotonic() + timeout if timeout is not None else None

        if self._unbuffered:
            return self._send_unbuffered(value, deadline, monotonic)
        return self._send_buffered(value, deadline, monotonic)

    def _send_buffered(
        self, value: T, deadline: float | None, monotonic: Callable[[], float]
    ) -> bool:
        buffer = self._buffer
        buffer_size = self._buffer_size
        not_full = self._not_full

        with not_full:
            if self._closed:
                raise ValueError("send on closed channel")
            while len(buffer) >= buffer_size:
                if self._closed:
                    raise ValueError("send on closed channel")
                remaining = None
                if deadline is not None:
                    remaining = deadline - monotonic()
                    if remaining <= 0:
                        return False
                if not not_full.wait(remaining):
                    return False
            buffer.append(value)
            self._not_empty.notify()
            _notify_select_waiters()
            return True

    def _send_unbuffered(
        self, value: T, deadline: float | None, monotonic: Callable[[], float]
    ) -> bool:
        """True unbuffered send: block until a receiver takes the value."""
        with self._lock:
            if self._closed:
                raise ValueError("send on closed channel")
            # Wait for any previous rendezvous to complete
            while self._rendezvous_value is not _CHAN_CLOSED and not self._rendezvous_taken:
                self._rendezvous.wait(0.01)
            # Place value and wait for receiver to take it
            self._rendezvous_value = value
            self._rendezvous_taken = False
            self._not_empty.notify()
            _notify_select_waiters()
            # Wait until receiver picks up the value
            while not self._rendezvous_taken:
                if self._closed:
                    raise ValueError("send on closed channel")
                remaining = None
                if deadline is not None:
                    remaining = deadline - monotonic()
                    if remaining <= 0:
                        self._rendezvous_value = _CHAN_CLOSED
                        return False
                self._rendezvous.wait(remaining)
            return True

    def Recv(self, timeout: float | None = None) -> tuple[T | None, bool]:
        """Receive a value from the channel.

        For unbuffered channels, completes the rendezvous with the sender.
        """
        monotonic = time.monotonic
        deadline = monotonic() + timeout if timeout is not None else None

        if self._unbuffered:
            return self._recv_unbuffered(deadline, monotonic)
        return self._recv_buffered(deadline, monotonic)

    def _recv_buffered(
        self, deadline: float | None, monotonic: Callable[[], float]
    ) -> tuple[T | None, bool]:
        buffer = self._buffer
        not_empty = self._not_empty

        with not_empty:
            while not buffer:
                if self._closed:
                    return None, False
                remaining = None
                if deadline is not None:
                    remaining = deadline - monotonic()
                    if remaining <= 0:
                        return None, False
                if not not_empty.wait(remaining):
                    if self._closed:
                        return None, False
                    return None, False
            value = buffer.popleft()
            self._not_full.notify()
            _notify_select_waiters()
            return value, True

    def _recv_unbuffered(
        self, deadline: float | None, monotonic: Callable[[], float]
    ) -> tuple[T | None, bool]:
        """True unbuffered recv: wait for sender, then take the value."""
        with self._lock:
            # Wait for a sender to place a value (not sentinel AND not yet taken)
            while True:
                if self._rendezvous_value is not _CHAN_CLOSED and not self._rendezvous_taken:
                    break  # A sender has placed a value, take it
                if self._closed:
                    return None, False
                remaining = None
                if deadline is not None:
                    remaining = deadline - monotonic()
                    if remaining <= 0:
                        return None, False
                self._not_empty.wait(remaining)
            # Take the value and notify sender
            value = cast(T, self._rendezvous_value)
            self._rendezvous_value = _CHAN_CLOSED
            self._rendezvous_taken = True
            self._rendezvous.notify()
            _notify_select_waiters()
            return value, True

    def TryRecv(self) -> tuple[T | None, bool]:
        """Non-blocking receive. Returns (None, False) if nothing available."""
        with self._lock:
            if self._unbuffered:
                if self._rendezvous_value is not _CHAN_CLOSED and not self._rendezvous_taken:
                    value = cast(T, self._rendezvous_value)
                    self._rendezvous_value = _CHAN_CLOSED
                    self._rendezvous_taken = True
                    self._rendezvous.notify()
                    _notify_select_waiters()
                    return value, True
                return None, False
            if self._buffer:
                value = self._buffer.popleft()
                self._not_full.notify()
                _notify_select_waiters()
                return value, True
            return None, False

    def TrySend(self, value: T) -> bool:
        """Non-blocking send. Returns False if buffer full or closed."""
        with self._lock:
            if self._closed:
                return False
            if self._unbuffered:
                # Can only succeed if a receiver is already waiting
                # For TrySend on unbuffered, we use the buffer as a handoff slot
                # This is inherently racy; return False (Go semantics: TrySend on
                # unbuffered only succeeds if a goroutine is blocked in Recv)
                return False
            if len(self._buffer) < self._buffer_size:
                self._buffer.append(value)
                self._not_empty.notify()
                _notify_select_waiters()
                return True
            return False

    def Close(self) -> None:
        """Close the channel. Sends will panic, receives drain then return (nil, false)."""
        with self._lock:
            if self._closed:
                raise ValueError("close of closed channel")
            self._closed = True
            self._not_empty.notify_all()
            self._not_full.notify_all()
            self._rendezvous.notify_all()
            _notify_select_waiters()

    @property
    def closed(self) -> bool:
        return self._closed

    def __iter__(self) -> Iterator[T]:
        """Iterate over channel values until closed and drained.

        Blocks properly on condition variables - no polling.
        """
        while True:
            val, ok = self.Recv()
            if not ok:
                break
            yield cast(T, val)

    def __len__(self) -> int:
        with self._lock:
            return len(self._buffer)


@dataclass
class SelectCase:
    """A case for the Select statement."""

    chan: Chan[Any]
    send_value: Any = None
    is_send: bool = False
    is_default: bool = False


def Select(*cases: SelectCase, default: bool = False) -> tuple[int, Any, bool]:
    """Select waits on multiple channel operations.

    Uses condition-variable notification instead of spin-wait for zero
    CPU waste and minimal latency.

    Args:
        *cases: SelectCase objects for each channel operation
        default: If True, don't block if no cases ready

    Returns:
        (index, value, ok) - index of selected case, received value, ok flag

    Example:
        ch1, ch2 = Chan[int](), Chan[str]()

        idx, val, ok = Select(
            SelectCase(ch1),           # Receive from ch1
            SelectCase(ch2),           # Receive from ch2
            default=True               # Don't block
        )

        if idx == 0:
            print(f"Got from ch1: {val}")
        elif idx == 1:
            print(f"Got from ch2: {val}")
        else:
            print("No channel ready")

    """
    indices = list(range(len(cases)))
    random.shuffle(indices)

    # First pass: try non-blocking
    for idx in indices:
        case = cases[idx]
        if case.is_send:
            if case.chan.TrySend(case.send_value):
                return idx, None, True
        else:
            val, ok = case.chan.TryRecv()
            if ok or case.chan.closed:
                return idx, val, ok

    if default:
        return -1, None, False

    # Blocking path: register a condition variable and wait for notification
    # from any channel involved in this Select.
    cond = threading.Condition()
    with _select_waiters_lock:
        _select_waiters.append(cond)

    try:
        with cond:
            while True:
                # Try all cases (randomized for fairness)
                random.shuffle(indices)
                for idx in indices:
                    case = cases[idx]
                    if case.is_send:
                        if case.chan.TrySend(case.send_value):
                            return idx, None, True
                    else:
                        val, ok = case.chan.TryRecv()
                        if ok or case.chan.closed:
                            return idx, val, ok
                # Wait for any channel to notify us
                cond.wait(timeout=0.1)
    finally:
        with _select_waiters_lock:
            if cond in _select_waiters:
                _select_waiters.remove(cond)


class Mutex:
    """A mutual exclusion lock (like sync.Mutex in Go)."""

    __slots__ = ("_lock",)

    def __init__(self) -> None:
        self._lock = threading.Lock()

    def Lock(self) -> None:
        self._lock.acquire()

    def Unlock(self) -> None:
        self._lock.release()

    def TryLock(self) -> bool:
        return self._lock.acquire(blocking=False)

    def __enter__(self) -> Mutex:
        self._lock.acquire()
        return self

    def __exit__(self, *args: object) -> None:
        self._lock.release()


class RWMutex:
    """A reader/writer mutual exclusion lock (like sync.RWMutex in Go).

    Supports concurrent reads but exclusive writes. Uses writer-preference
    to prevent writer starvation: when a writer is waiting, new readers
    block until the writer completes.
    """

    __slots__ = ("_state_lock", "_readers", "_writer_waiting", "_no_readers", "_write_lock")

    def __init__(self) -> None:
        self._state_lock = threading.Lock()
        self._readers = 0
        self._writer_waiting = False
        self._no_readers = threading.Event()
        self._no_readers.set()  # Initially no readers
        self._write_lock = threading.Lock()

    def Lock(self) -> None:
        """Acquire write lock. Blocks new readers and waits for existing ones."""
        self._write_lock.acquire()
        with self._state_lock:
            self._writer_waiting = True
        # Wait for all current readers to finish
        self._no_readers.wait()

    def Unlock(self) -> None:
        """Release write lock."""
        with self._state_lock:
            self._writer_waiting = False
        self._write_lock.release()

    def RLock(self) -> None:
        """Acquire read lock. Blocks if a writer is waiting or holding the lock."""
        # If a writer is waiting/holding, we must wait for it
        self._write_lock.acquire()
        with self._state_lock:
            self._readers += 1
            self._no_readers.clear()
        self._write_lock.release()

    def RUnlock(self) -> None:
        """Release read lock."""
        with self._state_lock:
            self._readers -= 1
            if self._readers == 0:
                self._no_readers.set()

    def TryLock(self) -> bool:
        """Try to acquire write lock without blocking."""
        if not self._write_lock.acquire(blocking=False):
            return False
        with self._state_lock:
            if self._readers > 0:
                self._write_lock.release()
                return False
            self._writer_waiting = True
        return True

    def TryRLock(self) -> bool:
        """Try to acquire read lock without blocking."""
        if not self._write_lock.acquire(blocking=False):
            return False
        with self._state_lock:
            self._readers += 1
            self._no_readers.clear()
        self._write_lock.release()
        return True

    def __enter__(self) -> RWMutex:
        self.Lock()
        return self

    def __exit__(self, *args: object) -> None:
        self.Unlock()

    def RLocker(self) -> _RLocker:
        """Return a context manager for read locking."""
        return _RLocker(self)


class _RLocker:
    __slots__ = ("_rw",)

    def __init__(self, rw: RWMutex) -> None:
        self._rw = rw

    def __enter__(self) -> _RLocker:
        self._rw.RLock()
        return self

    def __exit__(self, *args: object) -> None:
        self._rw.RUnlock()


class Once:
    """Once ensures a function is only executed once (like sync.Once in Go)."""

    __slots__ = ("_done", "_lock")

    def __init__(self) -> None:
        self._done = False
        self._lock = threading.Lock()

    def Do(self, fn: Callable[[], None]) -> None:
        """Execute fn if it hasn't been executed yet."""
        if self._done:
            return
        with self._lock:
            if not self._done:
                fn()
                self._done = True


class Pool(Generic[T]):
    """A pool of reusable objects (like sync.Pool in Go).

    Uses thread-local storage for a fast path: each thread keeps one
    cached object that can be Get/Put without locking. Overflow goes
    to a shared pool protected by a lock.
    """

    __slots__ = ("_new", "_shared", "_lock", "_local")

    def __init__(self, new: Callable[[], T] | None = None):
        self._new = new
        self._shared: list[T] = []
        self._lock = threading.Lock()
        self._local = threading.local()

    def Get(self) -> T | None:
        """Get an object from the pool. Thread-local fast path first."""
        # Fast path: check thread-local cache (no lock)
        local = self._local
        obj = getattr(local, "cached", None)
        if obj is not None:
            local.cached = None
            return obj
        # Slow path: shared pool
        with self._lock:
            if self._shared:
                return self._shared.pop()
        if self._new:
            return self._new()
        return None

    def Put(self, x: T) -> None:
        """Put an object back in the pool. Thread-local fast path first."""
        # Fast path: store in thread-local if empty
        local = self._local
        if getattr(local, "cached", None) is None:
            local.cached = x
            return
        # Slow path: shared pool
        with self._lock:
            self._shared.append(x)


class GoGroup:
    """Context manager for spawning goroutines with automatic wait.

    Example:
        with GoGroup() as g:
            for i in range(10):
                g.go(worker, i)
            # Can also get futures
            future = g.go(compute, 42)
        # Automatically waits for all goroutines here

        # With max concurrency limit
        with GoGroup(limit=4) as g:
            for url in urls:
                g.go(fetch, url)

        # FAST PATH: Use go_map for batch operations (matches TPE speed)
        with GoGroup() as g:
            results = g.go_map(process, items)  # Returns list of results

    """

    __slots__ = ("_executor", "_owned_executor", "_pending")

    def __init__(self, limit: int | None = None):
        from concurrent.futures import ThreadPoolExecutor

        if limit:
            self._executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=limit)
            self._owned_executor = True
        else:
            executor = _get_runtime_fast()._executor
            assert executor is not None
            self._executor = executor
            self._owned_executor = False
        self._pending: list[Future[Any]] = []

    @property
    def executor(self) -> Any:
        return self._executor

    def go(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> Future[Any]:
        executor = self._executor
        if not args and not kwargs:
            future = executor.submit(fn)
        elif not kwargs:
            future = executor.submit(fn, *args)
        else:
            future = executor.submit(fn, *args, **kwargs)
        if not self._owned_executor:
            self._pending.append(future)
        return future

    def go1(self, fn: Callable[[Any], Any], arg: Any) -> Future[Any]:
        future = self._executor.submit(fn, arg)
        if not self._owned_executor:
            self._pending.append(future)
        return future

    def go_map(self, fn: Callable[..., T], items: list[Any]) -> list[T]:
        return list(self._executor.map(fn, items))

    def go_batch(
        self, fn: Callable[..., Any], args_list: list[tuple[Any, ...]]
    ) -> list[Future[Any]]:
        """Submit multiple tasks at once. More efficient than individual go() calls."""
        executor = self._executor
        futures_list = [executor.submit(fn, *args) for args in args_list]
        if not self._owned_executor:
            self._pending.extend(futures_list)
        return futures_list

    def Wait(self, timeout: float | None = None) -> bool:
        if self._owned_executor:
            return True
        pending = self._pending
        if not pending:
            return True
        done, not_done = futures_wait(pending, timeout=timeout)
        return len(not_done) == 0

    def __enter__(self) -> GoGroup:
        return self

    def __exit__(self, *args: object) -> None:
        if self._owned_executor:
            self._executor.shutdown(wait=True)
        elif self._pending:
            futures_wait(self._pending)


class ErrGroup:
    """Like Go's errgroup - manages goroutines that can fail.

    If any goroutine returns an error (raises exception), it's captured.
    On exit, the first error is raised.

    Example:
        with ErrGroup() as g:
            g.go(fetch_user, user_id)
            g.go(fetch_posts, user_id)
            g.go(fetch_comments, user_id)
        # Raises first exception if any task failed

        # Or check error without raising
        g = ErrGroup()
        g.go(task1)
        g.go(task2)
        err = g.Wait()  # Returns exception or None
        if err:
            print(f"Failed: {err}")

    """

    __slots__ = ("_wg", "_first_error", "_lock", "_executor", "_owned_executor")

    def __init__(self, limit: int | None = None) -> None:
        from concurrent.futures import ThreadPoolExecutor

        self._wg = WaitGroup()
        self._first_error: BaseException | None = None
        self._lock = threading.Lock()
        if limit:
            self._executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=limit)
            self._owned_executor = True
        else:
            executor = _get_runtime_fast()._executor
            assert executor is not None
            self._executor = executor
            self._owned_executor = False

    def _error_callback(self, future: Future[Any]) -> None:
        try:
            future.result()
        except BaseException as e:
            with self._lock:
                if self._first_error is None:
                    self._first_error = e
        finally:
            self._wg.Add(-1)

    def go(self, fn: Callable[..., Any], *args: Any, **kwargs: Any) -> Future[Any]:
        self._wg.Add(1)

        executor = self._executor
        if not args and not kwargs:
            future = executor.submit(fn)
        elif not kwargs:
            future = executor.submit(fn, *args)
        else:
            future = executor.submit(fn, *args, **kwargs)

        future.add_done_callback(self._error_callback)
        return future

    def Wait(self, timeout: float | None = None) -> BaseException | None:
        """Wait for all goroutines. Returns first error or None."""
        self._wg.Wait(timeout)
        return self._first_error

    def __enter__(self) -> ErrGroup:
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: object
    ) -> None:
        if self._owned_executor:
            self._executor.shutdown(wait=True)
        else:
            self._wg.Wait()
        if self._first_error is not None and exc_val is None:
            raise self._first_error


def goroutine(fn: Callable[..., T]) -> Callable[..., Future[T]]:
    """Decorator to make a function always run as a goroutine.

    Example:
        @goroutine
        def fetch_data(url):
            return requests.get(url).json()

        # Now calls return Future immediately
        future = fetch_data("https://api.example.com")
        result = future.result()  # Wait for result

        # Fire and forget
        fetch_data("https://api.example.com")

    """

    @functools.wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Future[T]:
        return go(fn, *args, **kwargs)

    return wrapper


def Await(*futures: Future[T], timeout: float | None = None) -> list[T]:
    """Wait for multiple futures and return their results.

    Example:
        f1 = go(task1)
        f2 = go(task2)
        f3 = go(task3)

        results = Await(f1, f2, f3)
        # results = [result1, result2, result3]

    """
    return [f.result(timeout=timeout) for f in futures]


def AfterFunc(delay: float, fn: Callable[[], Any]) -> Callable[[], bool]:
    """Run fn after delay seconds. Returns a cancel function.

    Like Go's time.AfterFunc.

    Example:
        cancel = AfterFunc(5.0, lambda: print("5 seconds!"))

        # Can cancel before it fires
        cancelled = cancel()

    """
    cancelled = threading.Event()
    executor = _get_runtime_fast()._executor
    assert executor is not None

    def runner() -> None:
        if not cancelled.wait(delay):
            fn()

    executor.submit(runner)

    def cancel() -> bool:
        if cancelled.is_set():
            return False
        cancelled.set()
        return True

    return cancel


def Ticker(interval: float) -> Chan[float]:
    """Returns a channel that receives the current time at regular intervals.

    Like Go's time.Ticker.

    Example:
        ticker = Ticker(1.0)  # tick every second
        for t in ticker:
            print(f"Tick at {t}")
            if some_condition:
                ticker.Close()
                break

    """
    ch: Chan[float] = Chan[float](buffer=1)

    def tick() -> None:
        while not ch.closed:
            time.sleep(interval)
            if ch.closed:
                break
            ch.TrySend(time.time())

    go(tick)
    return ch


def After(delay: float) -> Chan[float]:
    """Returns a channel that receives the time after delay seconds.

    Like Go's time.After.

    Example:
        timeout = After(5.0)

        idx, val, ok = Select(
            SelectCase(data_chan),
            SelectCase(timeout),
        )
        if idx == 1:
            print("Timed out!")

    """
    ch: Chan[float] = Chan[float](buffer=1)

    def send() -> None:
        time.sleep(delay)
        ch.TrySend(time.time())
        ch.Close()

    go(send)
    return ch


class Semaphore:
    """A counting semaphore for limiting concurrent access (like a weighted sync primitive).

    Useful for rate limiting or bounding concurrency to N simultaneous operations.

    Example:
        sem = Semaphore(3)  # Allow 3 concurrent operations

        def worker(id):
            with sem:
                do_work(id)

        with GoGroup() as g:
            for i in range(100):
                g.go(worker, i)  # Only 3 run at a time

    """

    __slots__ = ("_value", "_lock", "_condition")

    def __init__(self, value: int = 1) -> None:
        if value < 0:
            raise ValueError("semaphore initial value must be >= 0")
        self._value = value
        self._lock = threading.Lock()
        self._condition = threading.Condition(self._lock)

    def Acquire(self, n: int = 1, timeout: float | None = None) -> bool:
        """Acquire n permits. Blocks until available or timeout."""
        monotonic = time.monotonic
        deadline = monotonic() + timeout if timeout is not None else None
        with self._condition:
            while self._value < n:
                remaining = None
                if deadline is not None:
                    remaining = deadline - monotonic()
                    if remaining <= 0:
                        return False
                if not self._condition.wait(remaining):
                    return False
            self._value -= n
            return True

    def Release(self, n: int = 1) -> None:
        """Release n permits."""
        with self._condition:
            self._value += n
            self._condition.notify_all()

    def TryAcquire(self, n: int = 1) -> bool:
        """Try to acquire n permits without blocking."""
        with self._lock:
            if self._value >= n:
                self._value -= n
                return True
            return False

    def __enter__(self) -> Semaphore:
        self.Acquire()
        return self

    def __exit__(self, *args: object) -> None:
        self.Release()


def pipe(
    src: Chan[T],
    fn: Callable[[T], Any],
    dst: Chan[Any] | None = None,
    buffer: int = 0,
    workers: int = 1,
) -> Chan[Any]:
    """Pipe values from src through fn into dst (created if not provided).

    Creates a processing pipeline stage. Each value from src is transformed
    by fn and sent to dst. Closes dst when src is drained.

    Args:
        src: Source channel
        fn: Transform function
        dst: Destination channel (created if None)
        buffer: Buffer size for created dst channel
        workers: Number of concurrent workers processing the pipe

    Returns:
        The destination channel

    Example:
        ch1 = Chan[int](buffer=10)
        ch2 = pipe(ch1, lambda x: x * 2)
        ch3 = pipe(ch2, lambda x: str(x))

        # Producer
        for i in range(10):
            ch1.Send(i)
        ch1.Close()

        # Consumer
        for val in ch3:
            print(val)  # "0", "2", "4", ...

    """
    if dst is None:
        dst = Chan[Any](buffer=buffer)

    out = dst

    def _worker() -> None:
        for val in src:
            result = fn(val)
            out.Send(result)

    def _run() -> None:
        if workers <= 1:
            _worker()
        else:
            wg = WaitGroup()
            for _ in range(workers):
                wg.Add(1)
                go(_worker, done=wg)
            wg.Wait()
        out.Close()

    go(_run)
    return out


def merge(*channels: Chan[T], buffer: int = 0) -> Chan[T]:
    """Fan-in: merge multiple channels into one.

    Values from all source channels are forwarded to the output channel.
    Output is closed when all sources are closed.

    Args:
        *channels: Source channels to merge
        buffer: Buffer size for the output channel

    Returns:
        Merged output channel

    Example:
        ch1 = Chan[int](buffer=5)
        ch2 = Chan[int](buffer=5)

        merged = merge(ch1, ch2)

        # Producers write to ch1 and ch2
        # Consumer reads from merged
        for val in merged:
            print(val)

    """
    out: Chan[T] = Chan[T](buffer=buffer)

    def _forward(src: Chan[T]) -> None:
        for val in src:
            out.Send(val)

    def _run() -> None:
        wg = WaitGroup()
        for ch in channels:
            wg.Add(1)
            go(_forward, ch, done=wg)
        wg.Wait()
        out.Close()

    go(_run)
    return out


def fan_out(
    src: Chan[T],
    n: int,
    buffer: int = 0,
) -> list[Chan[T]]:
    """Fan-out: distribute values from one channel to N output channels.

    Values are distributed round-robin across output channels.
    All outputs are closed when the source is drained.

    Args:
        src: Source channel
        n: Number of output channels
        buffer: Buffer size for each output channel

    Returns:
        List of output channels

    Example:
        jobs = Chan[int](buffer=100)
        worker_chans = fan_out(jobs, 4, buffer=10)

        # Each worker_chan gets ~1/4 of the jobs
        for i, ch in enumerate(worker_chans):
            go(worker, i, ch)

    """
    outputs: list[Chan[T]] = [Chan[T](buffer=buffer) for _ in range(n)]

    def _distribute() -> None:
        idx = 0
        for val in src:
            outputs[idx % n].Send(val)
            idx += 1
        for ch in outputs:
            ch.Close()

    go(_distribute)
    return outputs


def parallel_for(
    start: int, end: int, fn: Callable[[int], None], workers: int | None = None
) -> None:
    """Execute fn(i) for i in range(start, end) in parallel.

    Uses the M:N runtime for efficient scheduling with automatic chunking.

    Example:
        results = [None] * 100

        def process(i):
            results[i] = expensive_computation(i)

        parallel_for(0, 100, process)

    """
    runtime = _get_runtime_fast()
    # Use map with chunking for large ranges
    count = end - start
    chunksize = max(1, count // (runtime.num_workers * 4))
    list(runtime.map(fn, range(start, end), chunksize=chunksize))


def parallel_map(fn: Callable[[T], Any], items: list[T], workers: int | None = None) -> list[Any]:
    """Apply fn to each item in parallel, preserving order.

    Uses the M:N runtime for efficient scheduling with automatic chunking.

    Example:
        results = parallel_map(lambda x: x * 2, [1, 2, 3, 4, 5])
        # [2, 4, 6, 8, 10]

    """
    runtime = _get_runtime_fast()
    # Use map with chunking for large lists
    chunksize = max(1, len(items) // (runtime.num_workers * 4))
    return runtime.map(fn, items, chunksize=chunksize)



"""Tests for goated.std.binary module (encoding/binary)."""

from io import BytesIO

from goated.std import binary


class TestConstants:
    """Test binary constants."""

    def test_max_varint_len16(self):
        assert binary.MaxVarintLen16 == 3

    def test_max_varint_len32(self):
        assert binary.MaxVarintLen32 == 5

    def test_max_varint_len64(self):
        assert binary.MaxVarintLen64 == 10


class TestByteOrder:
    """Test ByteOrder class."""

    def test_big_endian_name(self):
        assert binary.BigEndian.String() == "BigEndian"

    def test_little_endian_name(self):
        assert binary.LittleEndian.String() == "LittleEndian"


class TestBigEndianUint16:
    """Test BigEndian Uint16 operations."""

    def test_uint16(self):
        b = bytes([0x01, 0x02])
        assert binary.BigEndian.Uint16(b) == 0x0102

    def test_put_uint16(self):
        b = bytearray(2)
        binary.BigEndian.PutUint16(b, 0x0102)
        assert b == bytearray([0x01, 0x02])

    def test_append_uint16(self):
        b = b"prefix"
        result = binary.BigEndian.AppendUint16(b, 0x0102)
        assert result == b"prefix\x01\x02"


class TestBigEndianUint32:
    """Test BigEndian Uint32 operations."""

    def test_uint32(self):
        b = bytes([0x01, 0x02, 0x03, 0x04])
        assert binary.BigEndian.Uint32(b) == 0x01020304

    def test_put_uint32(self):
        b = bytearray(4)
        binary.BigEndian.PutUint32(b, 0x01020304)
        assert b == bytearray([0x01, 0x02, 0x03, 0x04])

    def test_append_uint32(self):
        b = b""
        result = binary.BigEndian.AppendUint32(b, 0x01020304)
        assert result == bytes([0x01, 0x02, 0x03, 0x04])


class TestBigEndianUint64:
    """Test BigEndian Uint64 operations."""

    def test_uint64(self):
        b = bytes([0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08])
        assert binary.BigEndian.Uint64(b) == 0x0102030405060708

    def test_put_uint64(self):
        b = bytearray(8)
        binary.BigEndian.PutUint64(b, 0x0102030405060708)
        assert b == bytearray([0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08])

    def test_append_uint64(self):
        b = b""
        result = binary.BigEndian.AppendUint64(b, 256)
        assert len(result) == 8


class TestLittleEndianUint16:
    """Test LittleEndian Uint16 operations."""

    def test_uint16(self):
        b = bytes([0x02, 0x01])
        assert binary.LittleEndian.Uint16(b) == 0x0102

    def test_put_uint16(self):
        b = bytearray(2)
        binary.LittleEndian.PutUint16(b, 0x0102)
        assert b == bytearray([0x02, 0x01])


class TestLittleEndianUint32:
    """Test LittleEndian Uint32 operations."""

    def test_uint32(self):
        b = bytes([0x04, 0x03, 0x02, 0x01])
        assert binary.LittleEndian.Uint32(b) == 0x01020304

    def test_put_uint32(self):
        b = bytearray(4)
        binary.LittleEndian.PutUint32(b, 0x01020304)
        assert b == bytearray([0x04, 0x03, 0x02, 0x01])


class TestLittleEndianUint64:
    """Test LittleEndian Uint64 operations."""

    def test_uint64(self):
        b = bytes([0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01])
        assert binary.LittleEndian.Uint64(b) == 0x0102030405060708

    def test_put_uint64(self):
        b = bytearray(8)
        binary.LittleEndian.PutUint64(b, 0x0102030405060708)
        assert b == bytearray([0x08, 0x07, 0x06, 0x05, 0x04, 0x03, 0x02, 0x01])


class TestWrite:
    """Test Write function."""

    def test_write_bytes(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, b"hello")
        assert result.is_ok()
        assert buf.getvalue() == b"hello"

    def test_write_small_int(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, 42)
        assert result.is_ok()
        assert buf.getvalue() == bytes([42])

    def test_write_uint16_int(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, 0x0102)
        assert result.is_ok()
        assert buf.getvalue() == bytes([0x01, 0x02])

    def test_write_uint32_int(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, 0x01020304)
        assert result.is_ok()
        assert buf.getvalue() == bytes([0x01, 0x02, 0x03, 0x04])

    def test_write_float(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, 3.14)
        assert result.is_ok()
        assert len(buf.getvalue()) == 8

    def test_write_list(self):
        buf = BytesIO()
        result = binary.Write(buf, binary.BigEndian, [1, 2, 3])
        assert result.is_ok()
        assert len(buf.getvalue()) == 3


class TestRead:
    """Test Read function."""

    def test_read_list(self):
        buf = BytesIO(bytes(8 * 2))
        data = [0, 0]
        result = binary.Read(buf, binary.BigEndian, data)
        assert result.is_ok()

    def test_read_eof(self):
        buf = BytesIO(b"")
        data = [0]
        result = binary.Read(buf, binary.BigEndian, data)
        assert result.is_err()
        assert result.err().go_type == "io.EOF"


class TestSize:
    """Test Size function."""

    def test_size_bytes(self):
        assert binary.Size(b"hello") == 5

    def test_size_small_int(self):
        assert binary.Size(42) == 1

    def test_size_uint16_int(self):
        assert binary.Size(0x0102) == 2

    def test_size_uint32_int(self):
        assert binary.Size(0x01020304) == 4

    def test_size_large_int(self):
        assert binary.Size(0x0102030405060708) == 8

    def test_size_negative_int(self):
        assert binary.Size(-1) == 8

    def test_size_float(self):
        assert binary.Size(3.14) == 8

    def test_size_list(self):
        assert binary.Size([1, 2, 3]) == 3

    def test_size_empty(self):
        assert binary.Size(b"") == 0


class TestPutUvarint:
    """Test PutUvarint function."""

    def test_put_uvarint_small(self):
        buf = bytearray(10)
        n = binary.PutUvarint(buf, 1)
        assert n == 1
        assert buf[0] == 1

    def test_put_uvarint_127(self):
        buf = bytearray(10)
        n = binary.PutUvarint(buf, 127)
        assert n == 1
        assert buf[0] == 127

    def test_put_uvarint_128(self):
        buf = bytearray(10)
        n = binary.PutUvarint(buf, 128)
        assert n == 2
        assert buf[0] == 0x80
        assert buf[1] == 1

    def test_put_uvarint_300(self):
        buf = bytearray(10)
        n = binary.PutUvarint(buf, 300)
        assert n == 2


class TestUvarint:
    """Test Uvarint function."""

    def test_uvarint_small(self):
        buf = bytes([1])
        val, n = binary.Uvarint(buf)
        assert val == 1
        assert n == 1

    def test_uvarint_127(self):
        buf = bytes([127])
        val, n = binary.Uvarint(buf)
        assert val == 127
        assert n == 1

    def test_uvarint_128(self):
        buf = bytes([0x80, 0x01])
        val, n = binary.Uvarint(buf)
        assert val == 128
        assert n == 2

    def test_uvarint_empty(self):
        buf = bytes([])
        val, n = binary.Uvarint(buf)
        assert val == 0
        assert n == 0


class TestPutVarint:
    """Test PutVarint function."""

    def test_put_varint_positive(self):
        buf = bytearray(10)
        n = binary.PutVarint(buf, 1)
        assert n >= 1

    def test_put_varint_negative(self):
        buf = bytearray(10)
        n = binary.PutVarint(buf, -1)
        assert n >= 1

    def test_put_varint_zero(self):
        buf = bytearray(10)
        n = binary.PutVarint(buf, 0)
        assert n == 1
        assert buf[0] == 0


class TestVarint:
    """Test Varint function."""

    def test_varint_positive(self):
        buf = bytearray(10)
        binary.PutVarint(buf, 42)
        val, n = binary.Varint(bytes(buf))
        assert val == 42

    def test_varint_negative(self):
        buf = bytearray(10)
        binary.PutVarint(buf, -42)
        val, n = binary.Varint(bytes(buf))
        assert val == -42

    def test_varint_zero(self):
        buf = bytes([0])
        val, n = binary.Varint(buf)
        assert val == 0
        assert n == 1


class TestReadUvarint:
    """Test ReadUvarint function."""

    def test_read_uvarint(self):
        buf = BytesIO(bytes([128, 1]))
        result = binary.ReadUvarint(buf)
        assert result.is_ok()
        assert result.unwrap() == 128

    def test_read_uvarint_eof(self):
        buf = BytesIO(b"")
        result = binary.ReadUvarint(buf)
        assert result.is_err()
        assert result.err().go_type == "io.EOF"


class TestReadVarint:
    """Test ReadVarint function."""

    def test_read_varint(self):
        buf = bytearray(10)
        binary.PutVarint(buf, -100)
        reader = BytesIO(bytes(buf))

        result = binary.ReadVarint(reader)
        assert result.is_ok()
        assert result.unwrap() == -100


class TestRoundTrip:
    """Test roundtrip encoding/decoding."""

    def test_uvarint_roundtrip(self):
        for val in [0, 1, 127, 128, 255, 256, 65535, 2**20]:
            buf = bytearray(10)
            n = binary.PutUvarint(buf, val)
            decoded, m = binary.Uvarint(bytes(buf[:n]))
            assert decoded == val
            assert n == m

    def test_varint_roundtrip(self):
        for val in [0, 1, -1, 127, -128, 1000, -1000]:
            buf = bytearray(10)
            n = binary.PutVarint(buf, val)
            decoded, m = binary.Varint(bytes(buf[:n]))
            assert decoded == val


class TestModuleExports:
    """Test module exports."""

    def test_exports(self):
        assert hasattr(binary, "Read")
        assert hasattr(binary, "Write")
        assert hasattr(binary, "Size")
        assert hasattr(binary, "PutVarint")
        assert hasattr(binary, "PutUvarint")
        assert hasattr(binary, "Varint")
        assert hasattr(binary, "Uvarint")
        assert hasattr(binary, "ReadVarint")
        assert hasattr(binary, "ReadUvarint")
        assert hasattr(binary, "BigEndian")
        assert hasattr(binary, "LittleEndian")
        assert hasattr(binary, "ByteOrder")
        assert hasattr(binary, "MaxVarintLen16")
        assert hasattr(binary, "MaxVarintLen32")
        assert hasattr(binary, "MaxVarintLen64")



import io
import sys

import pytest

from goated.std import log


class TestPrint:
    def test_print(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.Print("test message")
        output = buf.getvalue()
        assert "test message" in output

    def test_print_multiple_args(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.Print("hello", "world")
        output = buf.getvalue()
        assert "hello" in output
        assert "world" in output


class TestPrintf:
    def test_printf(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.Printf("value: %d", 42)
        output = buf.getvalue()
        assert "value: 42" in output

    def test_printf_multiple(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.Printf("%s: %d", "count", 10)
        output = buf.getvalue()
        assert "count: 10" in output


class TestPrintln:
    def test_println(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.Println("line message")
        output = buf.getvalue()
        assert "line message" in output
        assert output.endswith("\n")


class TestPanic:
    def test_panic(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        with pytest.raises(RuntimeError) as exc_info:
            log.Panic("panic message")
        assert "panic message" in str(exc_info.value)

    def test_panicf(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        with pytest.raises(RuntimeError) as exc_info:
            log.Panicf("error code: %d", 500)
        assert "error code: 500" in str(exc_info.value)

    def test_panicln(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        with pytest.raises(RuntimeError) as exc_info:
            log.Panicln("panic", "line")
        assert "panic" in str(exc_info.value)


class TestSetPrefix:
    def test_set_prefix(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.SetPrefix("[TEST] ")
        log.Print("message")
        output = buf.getvalue()
        assert "[TEST]" in output

    def test_prefix(self):
        log.SetPrefix("[PREFIX] ")
        assert log.Prefix() == "[PREFIX] "
        log.SetPrefix("")


class TestSetFlags:
    def test_set_flags(self):
        log.SetFlags(log.Ldate | log.Ltime)
        assert log.Flags() == (log.Ldate | log.Ltime)

    def test_flags_date(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.SetFlags(log.Ldate)
        log.SetPrefix("")
        log.Print("test")
        output = buf.getvalue()
        assert "/" in output


class TestNew:
    def test_new_logger(self):
        buf = io.StringIO()
        logger = log.New(buf, "[CUSTOM] ", log.LstdFlags)

        logger.Print("custom message")
        output = buf.getvalue()
        assert "[CUSTOM]" in output
        assert "custom message" in output

    def test_new_logger_methods(self):
        buf = io.StringIO()
        logger = log.New(buf, "", 0)

        logger.Print("print")
        logger.Printf("printf %d", 1)
        logger.Println("println")

        output = buf.getvalue()
        assert "print" in output
        assert "printf 1" in output
        assert "println" in output


class TestLoggerSetters:
    def test_logger_set_output(self):
        buf1 = io.StringIO()
        buf2 = io.StringIO()
        logger = log.New(buf1, "", 0)

        logger.Print("first")
        logger.SetOutput(buf2)
        logger.Print("second")

        assert "first" in buf1.getvalue()
        assert "second" in buf2.getvalue()

    def test_logger_set_prefix(self):
        buf = io.StringIO()
        logger = log.New(buf, "", 0)

        logger.SetPrefix("[NEW] ")
        logger.Print("test")

        assert "[NEW]" in buf.getvalue()

    def test_logger_set_flags(self):
        buf = io.StringIO()
        logger = log.New(buf, "", 0)

        logger.SetFlags(log.Ldate)
        assert logger.Flags() == log.Ldate


class TestDefault:
    def test_default(self):
        default_logger = log.Default()
        assert default_logger is not None


class TestWriter:
    def test_writer(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        writer = log.Writer()
        assert writer is buf


class TestOutput:
    def test_output(self):
        buf = io.StringIO()
        log.SetOutput(buf)
        log.SetFlags(0)
        log.SetPrefix("")
        err = log.Output(1, "direct output")
        assert err is None
        assert "direct output" in buf.getvalue()


class TestConstants:
    def test_constants_defined(self):
        assert log.Ldate == 1
        assert log.Ltime == 2
        assert log.Lmicroseconds == 4
        assert log.Llongfile == 8
        assert log.Lshortfile == 16
        assert log.LUTC == 32
        assert log.Lmsgprefix == 64
        assert log.LstdFlags == (log.Ldate | log.Ltime)


class TestLoggerPanic:
    def test_logger_panic(self):
        buf = io.StringIO()
        logger = log.New(buf, "", 0)
        with pytest.raises(RuntimeError):
            logger.Panic("logger panic")

    def test_logger_panicf(self):
        buf = io.StringIO()
        logger = log.New(buf, "", 0)
        with pytest.raises(RuntimeError):
            logger.Panicf("logger panic %d", 1)


class TestCleanup:
    def test_reset_defaults(self):
        log.SetOutput(sys.stderr)
        log.SetPrefix("")
        log.SetFlags(log.LstdFlags)



"""Tests for Go path package bindings."""

from goated.std import path


class TestBase:
    def test_base_simple(self):
        assert path.Base("/a/b/c") == "c"

    def test_base_with_extension(self):
        assert path.Base("/home/user/file.txt") == "file.txt"

    def test_base_trailing_slash(self):
        assert path.Base("/a/b/c/") == "c"

    def test_base_single_element(self):
        assert path.Base("file.txt") == "file.txt"

    def test_base_root(self):
        assert path.Base("/") == "/"

    def test_base_empty(self):
        assert path.Base("") == "."

    def test_base_all_slashes(self):
        assert path.Base("///") == "/"


class TestClean:
    def test_clean_double_slash(self):
        assert path.Clean("a//b") == "a/b"

    def test_clean_dot(self):
        assert path.Clean("a/./b") == "a/b"

    def test_clean_dotdot(self):
        assert path.Clean("a/b/../c") == "a/c"

    def test_clean_rooted_dotdot(self):
        assert path.Clean("/../a") == "/a"

    def test_clean_complex(self):
        assert path.Clean("/a/b/c/./../../g") == "/a/g"

    def test_clean_empty(self):
        assert path.Clean("") == "."

    def test_clean_already_clean(self):
        assert path.Clean("/a/b/c") == "/a/b/c"

    def test_clean_trailing_slash(self):
        assert path.Clean("/a/b/c/") == "/a/b/c"

    def test_clean_root(self):
        assert path.Clean("/") == "/"


class TestDir:
    def test_dir_simple(self):
        assert path.Dir("/a/b/c") == "/a/b"

    def test_dir_file(self):
        assert path.Dir("/home/user/file.txt") == "/home/user"

    def test_dir_single_file(self):
        assert path.Dir("file.txt") == "."

    def test_dir_root(self):
        assert path.Dir("/") == "/"

    def test_dir_empty(self):
        assert path.Dir("") == "."

    def test_dir_trailing_slash(self):
        # Go's path.Dir includes the last element when there's a trailing slash
        assert path.Dir("/a/b/c/") == "/a/b/c"


class TestExt:
    def test_ext_simple(self):
        assert path.Ext("file.txt") == ".txt"

    def test_ext_multiple_dots(self):
        assert path.Ext("archive.tar.gz") == ".gz"

    def test_ext_no_extension(self):
        assert path.Ext("README") == ""

    def test_ext_hidden_file(self):
        # Go considers everything after first dot as extension, including ".gitignore"
        assert path.Ext(".gitignore") == ".gitignore"

    def test_ext_path_with_extension(self):
        assert path.Ext("/home/user/document.pdf") == ".pdf"

    def test_ext_empty(self):
        assert path.Ext("") == ""

    def test_ext_dot_only(self):
        assert path.Ext("file.") == "."


class TestIsAbs:
    def test_isabs_absolute(self):
        assert path.IsAbs("/home/user") is True

    def test_isabs_relative(self):
        assert path.IsAbs("home/user") is False

    def test_isabs_dot(self):
        assert path.IsAbs("./file") is False

    def test_isabs_dotdot(self):
        assert path.IsAbs("../file") is False

    def test_isabs_root(self):
        assert path.IsAbs("/") is True

    def test_isabs_empty(self):
        assert path.IsAbs("") is False


class TestMatch:
    def test_match_star(self):
        result = path.Match("*.txt", "file.txt")
        assert result.is_ok()
        assert result.unwrap() is True

    def test_match_star_no_match(self):
        result = path.Match("*.txt", "file.pdf")
        assert result.is_ok()
        assert result.unwrap() is False

    def test_match_question(self):
        result = path.Match("file?.txt", "file1.txt")
        assert result.is_ok()
        assert result.unwrap() is True

    def test_match_question_no_match(self):
        result = path.Match("file?.txt", "file10.txt")
        assert result.is_ok()
        assert result.unwrap() is False

    def test_match_bracket(self):
        result = path.Match("file[0-9].txt", "file5.txt")
        assert result.is_ok()
        assert result.unwrap() is True

    def test_match_bracket_no_match(self):
        result = path.Match("file[0-9].txt", "filea.txt")
        assert result.is_ok()
        assert result.unwrap() is False

    def test_match_exact(self):
        result = path.Match("file.txt", "file.txt")
        assert result.is_ok()
        assert result.unwrap() is True

    def test_match_bad_pattern(self):
        result = path.Match("[", "file")
        assert result.is_err()

    def test_match_complex(self):
        result = path.Match("*/*/*.txt", "a/b/c.txt")
        assert result.is_ok()
        assert result.unwrap() is True



from __future__ import annotations

import random as _random
from collections.abc import Callable
from typing import TypeVar

__all__ = [
    "Seed",
    "Int",
    "Intn",
    "Int31",
    "Int31n",
    "Int63",
    "Int63n",
    "Uint32",
    "Uint64",
    "Float32",
    "Float64",
    "NormFloat64",
    "ExpFloat64",
    "Perm",
    "Shuffle",
    "Read",
    "NewSource",
    "New",
    "Rand",
    "Source",
]

_rng = _random.Random()


def Seed(seed: int) -> None:
    """Seed uses the provided seed value to initialize the default Source."""
    _rng.seed(seed)


def Int() -> int:
    """Int returns a non-negative pseudo-random int."""
    return _rng.randint(0, (1 << 63) - 1)


def Intn(n: int) -> int:
    """Intn returns a non-negative pseudo-random number in [0,n)."""
    if n <= 0:
        raise ValueError("invalid argument to Intn")
    return _rng.randint(0, n - 1)


def Int31() -> int:
    """Int31 returns a non-negative pseudo-random 31-bit integer."""
    return _rng.randint(0, (1 << 31) - 1)


def Int31n(n: int) -> int:
    """Int31n returns a non-negative pseudo-random number in [0,n)."""
    if n <= 0:
        raise ValueError("invalid argument to Int31n")
    return _rng.randint(0, min(n - 1, (1 << 31) - 1))


def Int63() -> int:
    """Int63 returns a non-negative pseudo-random 63-bit integer."""
    return _rng.randint(0, (1 << 63) - 1)


def Int63n(n: int) -> int:
    """Int63n returns a non-negative pseudo-random number in [0,n)."""
    if n <= 0:
        raise ValueError("invalid argument to Int63n")
    return _rng.randint(0, n - 1)


def Uint32() -> int:
    """Uint32 returns a pseudo-random 32-bit value."""
    return _rng.randint(0, (1 << 32) - 1)


def Uint64() -> int:
    """Uint64 returns a pseudo-random 64-bit value."""
    return _rng.randint(0, (1 << 64) - 1)


def Float32() -> float:
    """Float32 returns a pseudo-random number in [0.0,1.0)."""
    return _rng.random()


def Float64() -> float:
    """Float64 returns a pseudo-random number in [0.0,1.0)."""
    return _rng.random()


def NormFloat64() -> float:
    """NormFloat64 returns a normally distributed float64 in [-inf, +inf]."""
    return _rng.gauss(0, 1)


def ExpFloat64() -> float:
    """ExpFloat64 returns an exponentially distributed float64 in (0, +inf]."""
    return _rng.expovariate(1)


T = TypeVar("T")


def Perm(n: int) -> list[int]:
    """Perm returns a pseudo-random permutation of the integers [0,n)."""
    result = list(range(n))
    _rng.shuffle(result)
    return result


def Shuffle(n: int, swap: Callable[[int, int], None]) -> None:
    """Shuffle pseudo-randomizes the order of elements using swap func."""
    for i in range(n - 1, 0, -1):
        j = _rng.randint(0, i)
        swap(i, j)


def Read(p: bytearray) -> tuple[int, None]:
    """Read generates len(p) random bytes and writes them into p."""
    for i in range(len(p)):
        p[i] = _rng.randint(0, 255)
    return len(p), None


class Source:
    """Source represents a source of uniformly-distributed pseudo-random int64 values."""

    def __init__(self, seed: int = 0):
        self._rng = _random.Random(seed)

    def Int63(self) -> int:
        return self._rng.randint(0, (1 << 63) - 1)

    def Seed(self, seed: int) -> None:
        self._rng.seed(seed)


class Rand:
    """Rand is a source of random numbers."""

    def __init__(self, src: Source):
        self._src = src
        self._rng = src._rng

    def Int(self) -> int:
        return self._rng.randint(0, (1 << 63) - 1)

    def Intn(self, n: int) -> int:
        if n <= 0:
            raise ValueError("invalid argument to Intn")
        return self._rng.randint(0, n - 1)

    def Int31(self) -> int:
        return self._rng.randint(0, (1 << 31) - 1)

    def Int31n(self, n: int) -> int:
        if n <= 0:
            raise ValueError("invalid argument to Int31n")
        return self._rng.randint(0, min(n - 1, (1 << 31) - 1))

    def Int63(self) -> int:
        return self._rng.randint(0, (1 << 63) - 1)

    def Int63n(self, n: int) -> int:
        if n <= 0:
            raise ValueError("invalid argument to Int63n")
        return self._rng.randint(0, n - 1)

    def Uint32(self) -> int:
        return self._rng.randint(0, (1 << 32) - 1)

    def Uint64(self) -> int:
        return self._rng.randint(0, (1 << 64) - 1)

    def Float32(self) -> float:
        return self._rng.random()

    def Float64(self) -> float:
        return self._rng.random()

    def NormFloat64(self) -> float:
        return self._rng.gauss(0, 1)

    def ExpFloat64(self) -> float:
        return self._rng.expovariate(1)

    def Perm(self, n: int) -> list[int]:
        result = list(range(n))
        self._rng.shuffle(result)
        return result

    def Shuffle(self, n: int, swap: Callable[[int, int], None]) -> None:
        for i in range(n - 1, 0, -1):
            j = self._rng.randint(0, i)
            swap(i, j)

    def Read(self, p: bytearray) -> tuple[int, None]:
        for i in range(len(p)):
            p[i] = self._rng.randint(0, 255)
        return len(p), None

    def Seed(self, seed: int) -> None:
        self._src.Seed(seed)


def NewSource(seed: int) -> Source:
    """NewSource returns a new pseudo-random Source seeded with the given value."""
    return Source(seed)


def New(src: Source) -> Rand:
    """New returns a new Rand that uses random values from src."""
    return Rand(src)



"""Drop-in replacement for Python's struct module.

Python's struct format strings are very Python-specific and the C-accelerated
implementation is already fast. This module is provided for completeness of
the compat namespace and passes all operations through to Python's struct.

Usage:
    from goated.compat import struct
    packed = struct.pack(">I", 42)
    (value,) = struct.unpack(">I", packed)
"""

from __future__ import annotations

# Re-export everything from Python's struct module
from struct import (  # noqa: F401
    Struct,
    calcsize,
    error,
    iter_unpack,
    pack,
    pack_into,
    unpack,
    unpack_from,
)

__all__ = [
    "pack",
    "pack_into",
    "unpack",
    "unpack_from",
    "iter_unpack",
    "calcsize",
    "error",
    "Struct",
]



"""Go type wrappers for Python.

This module provides Python classes that wrap Go types, enabling efficient
data exchange between Python and Go without unnecessary copying.

The key types are:
- GoSlice: Wraps Go slices ([]T)
- GoString: Wraps Go strings
- GoMap: Wraps Go maps (map[K]V)

These types integrate with ctypes for FFI and provide Pythonic interfaces.

Example:
    >>> from goated.types import GoSlice
    >>>
    >>> # From Python list
    >>> s = GoSlice.from_list([1, 2, 3, 4, 5])
    >>> print(s.to_list())
    [1, 2, 3, 4, 5]
    >>>
    >>> # Efficient iteration without copying
    >>> for item in s:
    ...     print(item)

"""

from __future__ import annotations

import ctypes
from collections.abc import ItemsView, Iterator, KeysView, Mapping, Sequence, ValuesView
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
)

if TYPE_CHECKING:
    pass

__all__ = ["GoSlice", "GoString", "GoMap", "GoInt", "GoFloat64"]

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# Go's integer type (platform dependent, but we use int64 for safety)
GoInt = ctypes.c_int64
GoUint = ctypes.c_uint64
GoFloat64 = ctypes.c_double
GoFloat32 = ctypes.c_float
GoBool = ctypes.c_bool
GoByte = ctypes.c_uint8


class _GoSliceStruct(ctypes.Structure):
    """C struct representation of Go slice header.

    Matches: struct { void *data; GoInt len; GoInt cap; }
    """

    _fields_ = [
        ("data", ctypes.c_void_p),
        ("len", ctypes.c_int64),
        ("cap", ctypes.c_int64),
    ]


class _GoStringStruct(ctypes.Structure):
    """C struct representation of Go string header.

    Matches: struct { const char *p; GoInt n; }
    """

    _fields_ = [
        ("p", ctypes.c_char_p),
        ("n", ctypes.c_int64),
    ]


class GoSlice(Generic[T]):
    """Python wrapper for Go slices.

    Provides a Pythonic interface to Go slices with efficient memory handling.
    Supports iteration, indexing, and conversion to/from Python lists.

    The underlying data is managed by Go's runtime. This class holds a handle
    to the Go slice, not a copy of the data.

    Attributes:
        _handle: Internal handle to the Go slice
        _len: Length of the slice
        _cap: Capacity of the slice
        _elem_type: Type of elements (for ctypes)

    Example:
        >>> s = GoSlice.from_list([1, 2, 3])
        >>> len(s)
        3
        >>> s[0]
        1
        >>> list(s)
        [1, 2, 3]

    """

    __slots__ = ("_handle", "_len", "_cap", "_elem_type", "_data")

    def __init__(
        self,
        handle: int = 0,
        length: int = 0,
        capacity: int = 0,
        elem_type: type = ctypes.c_int64,
        data: Any = None,
    ) -> None:
        """Initialize a GoSlice.

        Usually created via from_list() or returned from Go functions.

        Args:
            handle: Go handle to the slice (0 for empty/local)
            length: Number of elements
            capacity: Capacity of the slice
            elem_type: ctypes type of elements
            data: Raw ctypes array (for locally created slices)

        """
        self._handle = handle
        self._len = length
        self._cap = capacity if capacity > 0 else length
        self._elem_type = elem_type
        self._data = data

    @classmethod
    def from_list(
        cls,
        items: Sequence[T],
        elem_type: type = ctypes.c_int64,
    ) -> GoSlice[T]:
        """Create a GoSlice from a Python sequence.

        Args:
            items: Python sequence to convert
            elem_type: ctypes type for elements (default: int64)

        Returns:
            A new GoSlice containing the items

        Example:
            >>> s = GoSlice.from_list([1, 2, 3, 4, 5])
            >>> s.to_list()
            [1, 2, 3, 4, 5]
            >>>
            >>> # With specific type
            >>> s = GoSlice.from_list([1.5, 2.5], elem_type=ctypes.c_double)

        """
        n = len(items)
        if n == 0:
            return cls(handle=0, length=0, capacity=0, elem_type=elem_type)

        # Create ctypes array (type: ignore for ctypes dynamic array creation)
        ArrayType = elem_type * n  # type: ignore[operator]
        data = ArrayType(*items)  # type: ignore[operator]

        return cls(
            handle=0,
            length=n,
            capacity=n,
            elem_type=elem_type,
            data=data,
        )

    @staticmethod
    def from_bytes(data: bytes) -> GoSlice[int]:
        """Create a GoSlice[byte] from Python bytes.

        Args:
            data: Python bytes object

        Returns:
            A GoSlice containing the bytes

        """
        n = len(data)
        ArrayType = ctypes.c_uint8 * n
        arr = ArrayType(*data)
        return GoSlice[int](
            handle=0,
            length=n,
            capacity=n,
            elem_type=ctypes.c_uint8,
            data=arr,
        )

    def to_list(self) -> list[T]:
        """Convert the GoSlice to a Python list.

        Returns:
            A new Python list containing copies of all elements

        Example:
            >>> s = GoSlice.from_list([1, 2, 3])
            >>> s.to_list()
            [1, 2, 3]

        """
        if self._len == 0:
            return []

        if self._data is not None:
            return list(self._data[: self._len])

        # For Go-managed slices, we need to fetch through FFI
        # This will be implemented when we have the Go library loaded
        raise NotImplementedError("Go-managed slice conversion not yet implemented")

    def to_bytes(self) -> bytes:
        """Convert a byte slice to Python bytes.

        Returns:
            Python bytes object

        Raises:
            TypeError: If slice is not a byte slice

        """
        if self._elem_type != ctypes.c_uint8:
            raise TypeError("to_bytes() only works on byte slices")

        if self._len == 0:
            return b""

        if self._data is not None:
            return bytes(self._data[: self._len])

        raise NotImplementedError("Go-managed slice conversion not yet implemented")

    def to_struct(self) -> _GoSliceStruct:
        """Get the ctypes struct representation for FFI calls.

        Returns:
            _GoSliceStruct ready for passing to Go

        """
        struct = _GoSliceStruct()
        if self._data is not None:
            struct.data = ctypes.cast(self._data, ctypes.c_void_p)
        else:
            struct.data = ctypes.c_void_p(0)
        struct.len = self._len
        struct.cap = self._cap
        return struct

    def __len__(self) -> int:
        """Return the length of the slice."""
        return self._len

    def __getitem__(self, index: int) -> T:
        """Get an element by index.

        Args:
            index: Index (supports negative indexing)

        Returns:
            The element at the given index

        Raises:
            IndexError: If index is out of bounds

        """
        if index < 0:
            index = self._len + index
        if index < 0 or index >= self._len:
            raise IndexError(f"index {index} out of range for slice of length {self._len}")

        if self._data is not None:
            return self._data[index]  # type: ignore[no-any-return]

        raise NotImplementedError("Go-managed slice access not yet implemented")

    def __iter__(self) -> Iterator[T]:
        """Iterate over slice elements."""
        for i in range(self._len):
            yield self[i]

    def __contains__(self, item: T) -> bool:
        """Check if item is in the slice."""
        return any(x == item for x in self)

    def __repr__(self) -> str:
        if self._len <= 10:
            items = self.to_list()
            return f"GoSlice({items})"
        else:
            first_items = [self[i] for i in range(5)]
            return f"GoSlice([{', '.join(map(str, first_items))}, ... ({self._len} items)])"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, GoSlice):
            return self.to_list() == other.to_list()
        if isinstance(other, (list, tuple)):
            return self.to_list() == list(other)
        return False

    @property
    def capacity(self) -> int:
        """Return the capacity of the slice."""
        return self._cap


class GoString:
    """Python wrapper for Go strings.

    Go strings are immutable sequences of bytes (usually UTF-8). This class
    provides efficient conversion between Python strings and Go strings.

    Attributes:
        _data: The string data
        _len: Length in bytes

    Example:
        >>> s = GoString.from_str("Hello, World!")
        >>> str(s)
        'Hello, World!'
        >>> len(s)
        13

    """

    __slots__ = ("_data", "_len", "_bytes")

    def __init__(self, data: bytes = b"", length: int | None = None) -> None:
        """Initialize a GoString.

        Args:
            data: UTF-8 encoded bytes
            length: Length in bytes (computed if not provided)

        """
        self._bytes = data
        self._len = length if length is not None else len(data)
        # Keep reference to prevent garbage collection
        self._data = ctypes.c_char_p(data)

    @classmethod
    def from_str(cls, s: str) -> GoString:
        """Create a GoString from a Python string.

        Args:
            s: Python string

        Returns:
            A new GoString

        Example:
            >>> gs = GoString.from_str("hello")
            >>> str(gs)
            'hello'

        """
        encoded = s.encode("utf-8")
        return cls(encoded, len(encoded))

    def to_str(self) -> str:
        """Convert to a Python string.

        Returns:
            Python string (decoded from UTF-8)

        """
        return self._bytes.decode("utf-8")

    def to_bytes(self) -> bytes:
        """Get the raw bytes.

        Returns:
            The underlying bytes

        """
        return self._bytes

    def to_struct(self) -> _GoStringStruct:
        """Get the ctypes struct representation for FFI calls.

        Returns:
            _GoStringStruct ready for passing to Go

        """
        struct = _GoStringStruct()
        struct.p = self._data
        struct.n = self._len
        return struct

    def __str__(self) -> str:
        return self.to_str()

    def __repr__(self) -> str:
        return f"GoString({self._bytes!r})"

    def __len__(self) -> int:
        """Return length in bytes."""
        return self._len

    def __eq__(self, other: object) -> bool:
        if isinstance(other, GoString):
            return self._bytes == other._bytes
        if isinstance(other, str):
            return self.to_str() == other
        if isinstance(other, bytes):
            return self._bytes == other
        return False

    def __hash__(self) -> int:
        return hash(self._bytes)


class GoMap(Generic[K, V]):
    """Python wrapper for Go maps.

    Provides a dict-like interface to Go maps. Supports iteration,
    membership testing, and conversion to Python dicts.

    Note: Go maps are not ordered (until Go 1.12+), so iteration order
    may differ from Python dicts.

    Attributes:
        _handle: Internal handle to the Go map

    Example:
        >>> m = GoMap.from_dict({"a": 1, "b": 2})
        >>> m["a"]
        1
        >>> dict(m)
        {'a': 1, 'b': 2}

    """

    __slots__ = ("_handle", "_dict", "_key_type", "_value_type")

    def __init__(
        self,
        handle: int = 0,
        data: dict[K, V] | None = None,
        key_type: type = str,
        value_type: type = int,
    ) -> None:
        """Initialize a GoMap.

        Args:
            handle: Go handle to the map (0 for local)
            data: Python dict for locally created maps
            key_type: Type of keys
            value_type: Type of values

        """
        self._handle = handle
        self._dict = data if data is not None else {}
        self._key_type = key_type
        self._value_type = value_type

    @classmethod
    def from_dict(
        cls,
        d: Mapping[K, V],
        key_type: type = str,
        value_type: type = int,
    ) -> GoMap[K, V]:
        """Create a GoMap from a Python dict.

        Args:
            d: Python dict to convert
            key_type: Type of keys
            value_type: Type of values

        Returns:
            A new GoMap

        """
        return cls(handle=0, data=dict(d), key_type=key_type, value_type=value_type)

    def to_dict(self) -> dict[K, V]:
        """Convert to a Python dict.

        Returns:
            A new Python dict with copies of all items

        """
        return dict(self._dict)

    def __getitem__(self, key: K) -> V:
        """Get value by key."""
        return self._dict[key]

    def __setitem__(self, key: K, value: V) -> None:
        """Set value for key."""
        self._dict[key] = value

    def __delitem__(self, key: K) -> None:
        """Delete key."""
        del self._dict[key]

    def __contains__(self, key: object) -> bool:
        """Check if key exists."""
        return key in self._dict

    def __len__(self) -> int:
        """Return number of items."""
        return len(self._dict)

    def __iter__(self) -> Iterator[K]:
        """Iterate over keys."""
        return iter(self._dict)

    def keys(self) -> KeysView[K]:
        """Return keys view."""
        return self._dict.keys()

    def values(self) -> ValuesView[V]:
        """Return values view."""
        return self._dict.values()

    def items(self) -> ItemsView[K, V]:
        """Return items view."""
        return self._dict.items()

    def get(self, key: K, default: V | None = None) -> V | None:
        """Get value with default."""
        return self._dict.get(key, default)

    def __repr__(self) -> str:
        return f"GoMap({self._dict!r})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, GoMap):
            return self._dict == other._dict
        if isinstance(other, dict):
            return self._dict == other
        return False



"""Go os package bindings - Operating system functionality.

This module provides Python bindings for Go's os package.

Example:
    >>> from goated.std import goos as os
    >>>
    >>> os.Getenv("HOME")
    '/home/user'
    >>> os.Getwd()
    Ok('/current/dir')

"""

from __future__ import annotations

import os as _os
import shutil
import stat
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import IO, Any

from goated.result import Err, GoError, Ok, Result

__all__ = [
    # Environment
    "Getenv",
    "LookupEnv",
    "Setenv",
    "Unsetenv",
    "Clearenv",
    "Environ",
    "ExpandEnv",
    "Getwd",
    "Chdir",
    "TempDir",
    "UserHomeDir",
    "UserCacheDir",
    "UserConfigDir",
    "Hostname",
    "Getpid",
    "Getppid",
    "Getuid",
    "Getgid",
    "Geteuid",
    "Getegid",
    # File operations
    "Create",
    "Open",
    "OpenFile",
    "ReadFile",
    "WriteFile",
    "Remove",
    "RemoveAll",
    "Rename",
    "Mkdir",
    "MkdirAll",
    "MkdirTemp",
    "CreateTemp",
    "Stat",
    "Lstat",
    "Chmod",
    "Chown",
    "Link",
    "Symlink",
    "Readlink",
    "ReadDir",
    "IsExist",
    "IsNotExist",
    "IsPermission",
    "Truncate",
    "SameFile",
    # Constants
    "O_RDONLY",
    "O_WRONLY",
    "O_RDWR",
    "O_APPEND",
    "O_CREATE",
    "O_EXCL",
    "O_SYNC",
    "O_TRUNC",
    "PathSeparator",
    "PathListSeparator",
    "DevNull",
    # Types
    "FileInfo",
    "FileMode",
    "DirEntry",
    "File",
    # Args
    "Args",
    "Exit",
    "Executable",
]

# Constants
O_RDONLY = _os.O_RDONLY
O_WRONLY = _os.O_WRONLY
O_RDWR = _os.O_RDWR
O_APPEND = _os.O_APPEND
O_CREATE = _os.O_CREAT
O_EXCL = _os.O_EXCL
O_SYNC = getattr(_os, "O_SYNC", 0)
O_TRUNC = _os.O_TRUNC

PathSeparator = _os.sep
PathListSeparator = _os.pathsep
DevNull = _os.devnull

# Args
Args = sys.argv.copy()


def Exit(code: int) -> None:
    """Causes the current program to exit with the given status code."""
    sys.exit(code)


def Executable() -> Result[str, GoError]:
    """Returns the path name for the executable that started the current process."""
    try:
        return Ok(sys.executable)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


# Environment functions
def Getenv(key: str) -> str:
    """Retrieves the value of the environment variable named by the key."""
    return _os.environ.get(key, "")


def LookupEnv(key: str) -> tuple[str, bool]:
    """Retrieves the value of the environment variable and reports whether it was present."""
    if key in _os.environ:
        return _os.environ[key], True
    return "", False


def Setenv(key: str, value: str) -> Result[None, GoError]:
    """Sets the value of the environment variable named by the key."""
    try:
        _os.environ[key] = value
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def Unsetenv(key: str) -> Result[None, GoError]:
    """Unsets a single environment variable."""
    try:
        _os.environ.pop(key, None)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def Clearenv() -> None:
    """Deletes all environment variables."""
    _os.environ.clear()


def Environ() -> list[str]:
    """Returns a copy of strings representing the environment."""
    return [f"{k}={v}" for k, v in _os.environ.items()]


def ExpandEnv(s: str) -> str:
    """Replaces ${var} or $var in the string."""
    return _os.path.expandvars(s)


def Getwd() -> Result[str, GoError]:
    """Returns a rooted path name corresponding to the current directory."""
    try:
        return Ok(_os.getcwd())
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Chdir(dir: str) -> Result[None, GoError]:
    """Changes the current working directory to the named directory."""
    try:
        _os.chdir(dir)
        return Ok(None)
    except FileNotFoundError:
        return Err(GoError(f"chdir {dir}: no such file or directory", "os.PathError"))
    except PermissionError:
        return Err(GoError(f"chdir {dir}: permission denied", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def TempDir() -> str:
    """Returns the default directory to use for temporary files."""
    return tempfile.gettempdir()


def UserHomeDir() -> Result[str, GoError]:
    """Returns the current user's home directory."""
    try:
        home = Path.home()
        return Ok(str(home))
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def UserCacheDir() -> Result[str, GoError]:
    """Returns the default root directory to use for user-specific cached data."""
    try:
        if sys.platform == "darwin":
            return Ok(str(Path.home() / "Library" / "Caches"))
        elif sys.platform == "win32":
            local_app_data = _os.environ.get("LOCALAPPDATA", "")
            if local_app_data:
                return Ok(local_app_data)
            return Err(GoError("LOCALAPPDATA not set", "os.Error"))
        else:
            xdg_cache = _os.environ.get("XDG_CACHE_HOME", "")
            if xdg_cache:
                return Ok(xdg_cache)
            return Ok(str(Path.home() / ".cache"))
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def UserConfigDir() -> Result[str, GoError]:
    """Returns the default root directory to use for user-specific configuration data."""
    try:
        if sys.platform == "darwin":
            return Ok(str(Path.home() / "Library" / "Application Support"))
        elif sys.platform == "win32":
            app_data = _os.environ.get("APPDATA", "")
            if app_data:
                return Ok(app_data)
            return Err(GoError("APPDATA not set", "os.Error"))
        else:
            xdg_config = _os.environ.get("XDG_CONFIG_HOME", "")
            if xdg_config:
                return Ok(xdg_config)
            return Ok(str(Path.home() / ".config"))
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def Hostname() -> Result[str, GoError]:
    """Returns the host name reported by the kernel."""
    try:
        import socket

        return Ok(socket.gethostname())
    except Exception as e:
        return Err(GoError(str(e), "os.Error"))


def Getpid() -> int:
    """Returns the process id of the caller."""
    return _os.getpid()


def Getppid() -> int:
    """Returns the process id of the caller's parent."""
    return _os.getppid()


def Getuid() -> int:
    """Returns the numeric user id of the caller."""
    return getattr(_os, "getuid", lambda: -1)()


def Getgid() -> int:
    """Returns the numeric group id of the caller."""
    return getattr(_os, "getgid", lambda: -1)()


def Geteuid() -> int:
    """Returns the numeric effective user id of the caller."""
    return getattr(_os, "geteuid", lambda: -1)()


def Getegid() -> int:
    """Returns the numeric effective group id of the caller."""
    return getattr(_os, "getegid", lambda: -1)()


# FileMode represents a file's mode and permission bits
class FileMode(int):
    """FileMode represents a file's mode and permission bits."""

    # File mode bits
    ModeDir = 1 << 31
    ModeAppend = 1 << 30
    ModeExclusive = 1 << 29
    ModeTemporary = 1 << 28
    ModeSymlink = 1 << 27
    ModeDevice = 1 << 26
    ModeNamedPipe = 1 << 25
    ModeSocket = 1 << 24
    ModeSetuid = 1 << 23
    ModeSetgid = 1 << 22
    ModeCharDevice = 1 << 21
    ModeSticky = 1 << 20
    ModeIrregular = 1 << 19
    ModeType = (
        ModeDir
        | ModeSymlink
        | ModeNamedPipe
        | ModeSocket
        | ModeDevice
        | ModeCharDevice
        | ModeIrregular
    )
    ModePerm = 0o777

    def IsDir(self) -> bool:
        """Reports whether m describes a directory."""
        return bool(self & FileMode.ModeDir)

    def IsRegular(self) -> bool:
        """Reports whether m describes a regular file."""
        return (self & FileMode.ModeType) == 0

    def Perm(self) -> FileMode:
        """Returns the Unix permission bits."""
        return FileMode(self & FileMode.ModePerm)

    def String(self) -> str:
        """Returns a string representation of the file mode."""
        return format(self & FileMode.ModePerm, "o")


@dataclass
class FileInfo:
    """FileInfo describes a file and is returned by Stat and Lstat."""

    _name: str
    _size: int
    _mode: FileMode
    _mod_time: float
    _is_dir: bool
    _sys: object = None

    def Name(self) -> str:
        """Returns the base name of the file."""
        return self._name

    def Size(self) -> int:
        """Returns the length in bytes for regular files."""
        return self._size

    def Mode(self) -> FileMode:
        """Returns the file mode bits."""
        return self._mode

    def ModTime(self) -> float:
        """Returns the modification time."""
        return self._mod_time

    def IsDir(self) -> bool:
        """Abbreviation for Mode().IsDir()."""
        return self._is_dir

    def Sys(self) -> object:
        """Returns underlying data source."""
        return self._sys


@dataclass
class DirEntry:
    """DirEntry is an entry read from a directory."""

    _name: str
    _is_dir: bool
    _type: FileMode
    _info: FileInfo | None = None

    def Name(self) -> str:
        """Returns the name of the file."""
        return self._name

    def IsDir(self) -> bool:
        """Reports whether the entry describes a directory."""
        return self._is_dir

    def Type(self) -> FileMode:
        """Returns the type bits for the entry."""
        return self._type

    def Info(self) -> Result[FileInfo, GoError]:
        """Returns the FileInfo for the file or subdirectory."""
        if self._info:
            return Ok(self._info)
        return Err(GoError("info not available", "os.PathError"))


class File:
    """File represents an open file descriptor."""

    def __init__(self, f: IO[bytes], name: str) -> None:
        self._file = f
        self._name = name

    def Name(self) -> str:
        """Returns the name of the file."""
        return self._name

    def Read(self, n: int = -1) -> Result[bytes, GoError]:
        """Reads up to n bytes from the file."""
        try:
            data = self._file.read(n)
            return Ok(data if isinstance(data, bytes) else data.encode())
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def ReadAt(self, n: int, off: int) -> Result[bytes, GoError]:
        """Reads n bytes from the file starting at byte offset off."""
        try:
            current = self._file.tell()
            self._file.seek(off)
            data = self._file.read(n)
            self._file.seek(current)
            return Ok(data if isinstance(data, bytes) else data.encode())
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Write(self, data: bytes | str) -> Result[int, GoError]:
        """Writes data to the file."""
        try:
            if isinstance(data, str):
                data = data.encode()
            return Ok(self._file.write(data))
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def WriteString(self, s: str) -> Result[int, GoError]:
        """Writes a string to the file."""
        return self.Write(s.encode())

    def WriteAt(self, data: bytes, off: int) -> Result[int, GoError]:
        """Writes data to the file at offset off."""
        try:
            current = self._file.tell()
            self._file.seek(off)
            n = self._file.write(data)
            self._file.seek(current)
            return Ok(n)
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Seek(self, offset: int, whence: int = 0) -> Result[int, GoError]:
        """Seeks to the offset."""
        try:
            return Ok(self._file.seek(offset, whence))
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Close(self) -> Result[None, GoError]:
        """Closes the file."""
        try:
            self._file.close()
            return Ok(None)
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Sync(self) -> Result[None, GoError]:
        """Commits the current contents of the file to stable storage."""
        try:
            self._file.flush()
            if hasattr(self._file, "fileno"):
                _os.fsync(self._file.fileno())
            return Ok(None)
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Stat(self) -> Result[FileInfo, GoError]:
        """Returns the FileInfo for the file."""
        try:
            st = _os.fstat(self._file.fileno())
            return Ok(_stat_to_fileinfo(self._name, st))
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Truncate(self, size: int) -> Result[None, GoError]:
        """Changes the size of the file."""
        try:
            self._file.truncate(size)
            return Ok(None)
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))

    def Fd(self) -> int:
        """Returns the integer Unix file descriptor."""
        return self._file.fileno()

    def __enter__(self) -> File:
        return self

    def __exit__(self, *args: Any) -> None:
        self.Close()


def _stat_to_fileinfo(name: str, st: _os.stat_result) -> FileInfo:
    """Convert os.stat_result to FileInfo."""
    mode = FileMode(st.st_mode & 0o777)
    if stat.S_ISDIR(st.st_mode):
        mode = FileMode(mode | FileMode.ModeDir)
    if stat.S_ISLNK(st.st_mode):
        mode = FileMode(mode | FileMode.ModeSymlink)

    return FileInfo(
        _name=_os.path.basename(name),
        _size=st.st_size,
        _mode=mode,
        _mod_time=st.st_mtime,
        _is_dir=stat.S_ISDIR(st.st_mode),
        _sys=st,
    )


def Create(name: str) -> Result[File, GoError]:
    """Creates or truncates the named file."""
    try:
        f = open(name, "wb+")
        return Ok(File(f, name))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Open(name: str) -> Result[File, GoError]:
    """Opens the named file for reading."""
    try:
        f = open(name, "rb")
        return Ok(File(f, name))
    except FileNotFoundError:
        return Err(GoError(f"open {name}: no such file or directory", "os.PathError"))
    except PermissionError:
        return Err(GoError(f"open {name}: permission denied", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def OpenFile(name: str, flag: int, perm: int) -> Result[File, GoError]:
    """Opens the named file with specified flag and permissions."""
    try:
        mode = ""
        if flag & O_RDWR:
            mode = "r+b"
        elif flag & O_WRONLY:
            mode = "wb"
        else:
            mode = "rb"

        if flag & O_CREATE:
            if flag & O_EXCL:
                mode = "xb" if flag & O_WRONLY else "x+b"
            elif flag & O_TRUNC:
                mode = "wb" if flag & O_WRONLY else "w+b"
            elif flag & O_APPEND:
                mode = "ab" if flag & O_WRONLY else "a+b"
        elif flag & O_APPEND:
            mode = "ab" if flag & O_WRONLY else "a+b"

        f = open(name, mode)
        return Ok(File(f, name))
    except FileExistsError:
        return Err(GoError(f"open {name}: file exists", "os.PathError"))
    except FileNotFoundError:
        return Err(GoError(f"open {name}: no such file or directory", "os.PathError"))
    except PermissionError:
        return Err(GoError(f"open {name}: permission denied", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def ReadFile(name: str) -> Result[bytes, GoError]:
    """Reads the named file and returns the contents."""
    try:
        with open(name, "rb") as f:
            return Ok(f.read())
    except FileNotFoundError:
        return Err(GoError(f"open {name}: no such file or directory", "os.PathError"))
    except PermissionError:
        return Err(GoError(f"open {name}: permission denied", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def WriteFile(name: str, data: bytes | str, perm: int = 0o644) -> Result[None, GoError]:
    """Writes data to the named file, creating it if necessary."""
    try:
        if isinstance(data, str):
            data = data.encode()
        with open(name, "wb") as f:
            f.write(data)
        _os.chmod(name, perm)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Remove(name: str) -> Result[None, GoError]:
    """Removes the named file or empty directory."""
    try:
        _os.remove(name)
        return Ok(None)
    except IsADirectoryError:
        try:
            _os.rmdir(name)
            return Ok(None)
        except Exception as e:
            return Err(GoError(str(e), "os.PathError"))
    except FileNotFoundError:
        return Err(GoError(f"remove {name}: no such file or directory", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def RemoveAll(path: str) -> Result[None, GoError]:
    """Removes path and any children it contains."""
    try:
        if _os.path.isdir(path):
            shutil.rmtree(path)
        elif _os.path.exists(path):
            _os.remove(path)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Rename(oldpath: str, newpath: str) -> Result[None, GoError]:
    """Renames (moves) oldpath to newpath."""
    try:
        _os.rename(oldpath, newpath)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.LinkError"))


def Mkdir(name: str, perm: int = 0o755) -> Result[None, GoError]:
    """Creates a new directory with the specified permission bits."""
    try:
        _os.mkdir(name, perm)
        return Ok(None)
    except FileExistsError:
        return Err(GoError(f"mkdir {name}: file exists", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def MkdirAll(path: str, perm: int = 0o755) -> Result[None, GoError]:
    """Creates a directory named path, along with any necessary parents."""
    try:
        _os.makedirs(path, perm, exist_ok=True)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def MkdirTemp(dir: str, pattern: str) -> Result[str, GoError]:
    """Creates a new temporary directory and returns its path."""
    try:
        return Ok(tempfile.mkdtemp(prefix=pattern, dir=dir or None))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def CreateTemp(dir: str, pattern: str) -> Result[File, GoError]:
    """Creates a new temporary file and returns an open file."""
    try:
        fd, name = tempfile.mkstemp(prefix=pattern, dir=dir or None)
        f = _os.fdopen(fd, "w+b")
        return Ok(File(f, name))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Stat(name: str) -> Result[FileInfo, GoError]:
    """Returns a FileInfo describing the named file."""
    try:
        st = _os.stat(name)
        return Ok(_stat_to_fileinfo(name, st))
    except FileNotFoundError:
        return Err(GoError(f"stat {name}: no such file or directory", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Lstat(name: str) -> Result[FileInfo, GoError]:
    """Like Stat but does not follow symbolic links."""
    try:
        st = _os.lstat(name)
        return Ok(_stat_to_fileinfo(name, st))
    except FileNotFoundError:
        return Err(GoError(f"lstat {name}: no such file or directory", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Chmod(name: str, mode: int) -> Result[None, GoError]:
    """Changes the mode of the named file."""
    try:
        _os.chmod(name, mode)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Chown(name: str, uid: int, gid: int) -> Result[None, GoError]:
    """Changes the numeric uid and gid of the named file."""
    try:
        _os.chown(name, uid, gid)
        return Ok(None)
    except AttributeError:
        return Err(GoError("chown not supported on this platform", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def Link(oldname: str, newname: str) -> Result[None, GoError]:
    """Creates newname as a hard link to oldname."""
    try:
        _os.link(oldname, newname)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.LinkError"))


def Symlink(oldname: str, newname: str) -> Result[None, GoError]:
    """Creates newname as a symbolic link to oldname."""
    try:
        _os.symlink(oldname, newname)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.LinkError"))


def Readlink(name: str) -> Result[str, GoError]:
    """Returns the destination of the named symbolic link."""
    try:
        return Ok(_os.readlink(name))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def ReadDir(name: str) -> Result[list[DirEntry], GoError]:
    """Reads the named directory, returning all its directory entries."""
    try:
        entries = []
        for entry in _os.scandir(name):
            is_dir = entry.is_dir(follow_symlinks=False)
            mode = FileMode.ModeDir if is_dir else FileMode(0)
            if entry.is_symlink():
                mode = FileMode(mode | FileMode.ModeSymlink)

            entries.append(
                DirEntry(
                    _name=entry.name,
                    _is_dir=is_dir,
                    _type=FileMode(mode),
                )
            )
        entries.sort(key=lambda e: e._name)
        return Ok(entries)
    except FileNotFoundError:
        return Err(GoError(f"open {name}: no such file or directory", "os.PathError"))
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def IsExist(err: GoError) -> bool:
    """Returns whether the error is known to report that a file or directory already exists."""
    return "exists" in str(err).lower()


def IsNotExist(err: GoError) -> bool:
    """Returns whether the error is known to report that a file or directory does not exist."""
    return "no such file" in str(err).lower() or "not exist" in str(err).lower()


def IsPermission(err: GoError) -> bool:
    """Returns whether the error is known to report that permission is denied."""
    return "permission denied" in str(err).lower()


def Truncate(name: str, size: int) -> Result[None, GoError]:
    """Changes the size of the named file."""
    try:
        _os.truncate(name, size)
        return Ok(None)
    except Exception as e:
        return Err(GoError(str(e), "os.PathError"))


def SameFile(fi1: FileInfo, fi2: FileInfo) -> bool:
    """Reports whether fi1 and fi2 describe the same file."""
    if fi1._sys and fi2._sys:
        s1: Any = fi1._sys
        s2: Any = fi2._sys
        return bool(s1.st_dev == s2.st_dev and s1.st_ino == s2.st_ino)
    return False



from __future__ import annotations

import contextlib
import http.client as _http_client
import http.server as _http_server
import ssl
import urllib.parse
import urllib.request
from collections.abc import Callable
from dataclasses import dataclass, field
from io import BytesIO
from typing import IO, Any

from goated.result import Err, GoError, Ok, Result

__all__ = [
    "Get",
    "Post",
    "PostForm",
    "Head",
    "NewRequest",
    "DefaultClient",
    "Client",
    "Request",
    "Response",
    "Header",
    "Cookie",
    "Server",
    "ListenAndServe",
    "HandleFunc",
    "Handle",
    "Handler",
    "HandlerFunc",
    "ServeMux",
    "FileServer",
    "NotFound",
    "Redirect",
    "Error",
    "StatusOK",
    "StatusCreated",
    "StatusAccepted",
    "StatusNoContent",
    "StatusMovedPermanently",
    "StatusFound",
    "StatusNotModified",
    "StatusBadRequest",
    "StatusUnauthorized",
    "StatusForbidden",
    "StatusNotFound",
    "StatusMethodNotAllowed",
    "StatusInternalServerError",
    "StatusBadGateway",
    "StatusServiceUnavailable",
    "MethodGet",
    "MethodPost",
    "MethodPut",
    "MethodDelete",
    "MethodHead",
    "MethodOptions",
    "MethodPatch",
]

StatusOK = 200
StatusCreated = 201
StatusAccepted = 202
StatusNoContent = 204
StatusMovedPermanently = 301
StatusFound = 302
StatusNotModified = 304
StatusBadRequest = 400
StatusUnauthorized = 401
StatusForbidden = 403
StatusNotFound = 404
StatusMethodNotAllowed = 405
StatusInternalServerError = 500
StatusBadGateway = 502
StatusServiceUnavailable = 503

MethodGet = "GET"
MethodPost = "POST"
MethodPut = "PUT"
MethodDelete = "DELETE"
MethodHead = "HEAD"
MethodOptions = "OPTIONS"
MethodPatch = "PATCH"


class Header(dict[str, Any]):
    """HTTP header map."""

    def Add(self, key: str, value: str) -> None:
        """Add adds the key, value pair to the header."""
        key = key.title()
        if key in self:
            if isinstance(self[key], list):
                self[key].append(value)
            else:
                self[key] = [self[key], value]
        else:
            self[key] = value

    def Set(self, key: str, value: str) -> None:
        """Set sets the header entry associated with key to value."""
        self[key.title()] = value

    def Get(self, key: str) -> str:
        """Get gets the first value associated with the given key."""
        val = self.get(key.title(), "")
        if isinstance(val, list):
            return str(val[0]) if val else ""
        return str(val)

    def Del(self, key: str) -> None:
        """Del deletes the values associated with key."""
        self.pop(key.title(), None)

    def Clone(self) -> Header:
        """Clone returns a copy of the header."""
        h = Header()
        h.update(self)
        return h


@dataclass
class Cookie:
    """HTTP cookie."""

    Name: str = ""
    Value: str = ""
    Path: str = ""
    Domain: str = ""
    Expires: float | None = None
    MaxAge: int = 0
    Secure: bool = False
    HttpOnly: bool = False
    SameSite: str = ""

    def String(self) -> str:
        """String returns the serialization of the cookie."""
        parts = [f"{self.Name}={self.Value}"]
        if self.Path:
            parts.append(f"Path={self.Path}")
        if self.Domain:
            parts.append(f"Domain={self.Domain}")
        if self.MaxAge > 0:
            parts.append(f"Max-Age={self.MaxAge}")
        if self.Secure:
            parts.append("Secure")
        if self.HttpOnly:
            parts.append("HttpOnly")
        if self.SameSite:
            parts.append(f"SameSite={self.SameSite}")
        return "; ".join(parts)


@dataclass
class Request:
    """HTTP request."""

    Method: str = "GET"
    URL: str = ""
    Header: Header = field(default_factory=Header)
    Body: bytes | None = None
    ContentLength: int = 0
    Host: str = ""
    Form: dict[str, list[str]] = field(default_factory=dict)
    PostForm: dict[str, list[str]] = field(default_factory=dict)

    def AddCookie(self, c: Cookie) -> None:
        """AddCookie adds a cookie to the request."""
        existing = self.Header.Get("Cookie")
        if existing:
            self.Header.Set("Cookie", f"{existing}; {c.Name}={c.Value}")
        else:
            self.Header.Set("Cookie", f"{c.Name}={c.Value}")

    def Cookie(self, name: str) -> Result[Cookie, GoError]:
        """Cookie returns the named cookie."""
        cookie_header = self.Header.Get("Cookie")
        if not cookie_header:
            return Err(GoError("http: named cookie not present", "http.ErrNoCookie"))

        for part in cookie_header.split(";"):
            part = part.strip()
            if "=" in part:
                n, v = part.split("=", 1)
                if n.strip() == name:
                    return Ok(Cookie(Name=n.strip(), Value=v.strip()))

        return Err(GoError("http: named cookie not present", "http.ErrNoCookie"))

    def FormValue(self, key: str) -> str:
        """FormValue returns the first value for the named component of the query."""
        if key in self.Form:
            vals = self.Form[key]
            return vals[0] if vals else ""
        return ""

    def PostFormValue(self, key: str) -> str:
        """PostFormValue returns the first value for the named component of the POST form."""
        if key in self.PostForm:
            vals = self.PostForm[key]
            return vals[0] if vals else ""
        return ""


@dataclass
class Response:
    """HTTP response."""

    Status: str = ""
    StatusCode: int = 0
    Header: Header = field(default_factory=Header)
    Body: IO[bytes] | None = None
    ContentLength: int = -1
    Request: Request | None = None

    def Read(self) -> Result[bytes, GoError]:
        """Read reads the response body."""
        if self.Body is None:
            return Ok(b"")
        try:
            return Ok(self.Body.read())
        except Exception as e:
            return Err(GoError(str(e), "http.Error"))

    def Close(self) -> Result[None, GoError]:
        """Close closes the response body."""
        if self.Body is not None:
            with contextlib.suppress(Exception):
                self.Body.close()
        return Ok(None)


class Client:
    """HTTP client."""

    def __init__(self, timeout: float = 30.0):
        self.Timeout = timeout
        self.Transport = None
        self.CheckRedirect = None
        self.Jar = None

    def Do(self, req: Request) -> Result[Response, GoError]:
        """Do sends an HTTP request and returns an HTTP response."""
        try:
            parsed = urllib.parse.urlparse(req.URL)

            conn: _http_client.HTTPConnection | _http_client.HTTPSConnection
            if parsed.scheme == "https":
                context = ssl.create_default_context()
                conn = _http_client.HTTPSConnection(
                    parsed.netloc, timeout=self.Timeout, context=context
                )
            else:
                conn = _http_client.HTTPConnection(parsed.netloc, timeout=self.Timeout)

            path = parsed.path or "/"
            if parsed.query:
                path = f"{path}?{parsed.query}"

            headers = dict(req.Header)

            conn.request(req.Method, path, body=req.Body, headers=headers)
            resp = conn.getresponse()

            response = Response(
                Status=f"{resp.status} {resp.reason}",
                StatusCode=resp.status,
                Header=Header(resp.getheaders()),
                Body=BytesIO(resp.read()),
                ContentLength=int(resp.getheader("Content-Length", -1)),
                Request=req,
            )

            conn.close()
            return Ok(response)
        except TimeoutError:
            return Err(GoError("request timeout", "http.ErrTimeout"))
        except Exception as e:
            return Err(GoError(str(e), "http.Error"))

    def Get(self, url: str) -> Result[Response, GoError]:
        """Get issues a GET to the specified URL."""
        req = Request(Method="GET", URL=url)
        return self.Do(req)

    def Post(self, url: str, contentType: str, body: bytes) -> Result[Response, GoError]:
        """Post issues a POST to the specified URL."""
        req = Request(
            Method="POST",
            URL=url,
            Body=body,
            Header=Header({"Content-Type": contentType}),
        )
        return self.Do(req)

    def PostForm(self, url: str, data: dict[str, list[str]]) -> Result[Response, GoError]:
        """PostForm issues a POST with form data."""
        encoded = urllib.parse.urlencode(data, doseq=True)
        return self.Post(url, "application/x-www-form-urlencoded", encoded.encode())

    def Head(self, url: str) -> Result[Response, GoError]:
        """Head issues a HEAD to the specified URL."""
        req = Request(Method="HEAD", URL=url)
        return self.Do(req)


DefaultClient = Client()


def Get(url: str) -> Result[Response, GoError]:
    """Get issues a GET to the specified URL."""
    return DefaultClient.Get(url)


def Post(url: str, contentType: str, body: bytes) -> Result[Response, GoError]:
    """Post issues a POST to the specified URL."""
    return DefaultClient.Post(url, contentType, body)


def PostForm(url: str, data: dict[str, list[str]]) -> Result[Response, GoError]:
    """PostForm issues a POST with form data."""
    return DefaultClient.PostForm(url, data)


def Head(url: str) -> Result[Response, GoError]:
    """Head issues a HEAD to the specified URL."""
    return DefaultClient.Head(url)


def NewRequest(method: str, url: str, body: bytes | None = None) -> Result[Request, GoError]:
    """NewRequest returns a new Request given a method, URL, and optional body."""
    try:
        req = Request(
            Method=method.upper(),
            URL=url,
            Body=body,
            ContentLength=len(body) if body else 0,
        )
        parsed = urllib.parse.urlparse(url)
        req.Host = parsed.netloc
        return Ok(req)
    except Exception as e:
        return Err(GoError(str(e), "http.Error"))


# Handler takes ResponseWriter (not Response) - this is the HTTP server handler
Handler = Callable[["ResponseWriter", Request], None]
HandlerFunc = Handler


class ResponseWriter:
    """ResponseWriter interface for writing HTTP responses."""

    def __init__(self, handler: _http_server.BaseHTTPRequestHandler):
        self._handler = handler
        self._headers: Header = Header()
        self._status_code = 200
        self._headers_written = False

    def Header(self) -> Header:
        """Header returns the header map."""
        return self._headers

    def Write(self, data: bytes) -> int:
        """Write writes the data to the connection."""
        if not self._headers_written:
            self.WriteHeader(self._status_code)
        self._handler.wfile.write(data)
        return len(data)

    def WriteHeader(self, statusCode: int) -> None:
        """WriteHeader sends an HTTP response header."""
        if self._headers_written:
            return
        self._status_code = statusCode
        self._handler.send_response(statusCode)
        for key, value in self._headers.items():
            if isinstance(value, list):
                for v in value:
                    self._handler.send_header(key, v)
            else:
                self._handler.send_header(key, value)
        self._handler.end_headers()
        self._headers_written = True


_handlers: dict[str, Handler] = {}


class _HTTPHandler(_http_server.BaseHTTPRequestHandler):
    def do_GET(self) -> None:
        self._handle_request()

    def do_POST(self) -> None:
        self._handle_request()

    def do_PUT(self) -> None:
        self._handle_request()

    def do_DELETE(self) -> None:
        self._handle_request()

    def do_HEAD(self) -> None:
        self._handle_request()

    def do_OPTIONS(self) -> None:
        self._handle_request()

    def do_PATCH(self) -> None:
        self._handle_request()

    def _handle_request(self) -> None:
        path = urllib.parse.urlparse(self.path).path

        handler = None
        for pattern, h in _handlers.items():
            if path == pattern or path.startswith(pattern.rstrip("/") + "/"):
                handler = h
                break

        if handler is None:
            self.send_error(404, "Not Found")
            return

        content_length = int(self.headers.get("Content-Length", 0))
        body = self.rfile.read(content_length) if content_length > 0 else None

        req = Request(
            Method=self.command,
            URL=self.path,
            Header=Header(self.headers.items()),
            Body=body,
            Host=self.headers.get("Host", ""),
        )

        w = ResponseWriter(self)
        try:
            handler(w, req)
        except Exception as e:
            if not w._headers_written:
                self.send_error(500, str(e))

    def log_message(self, format: str, *args: Any) -> None:
        pass


class ServeMux:
    """ServeMux is an HTTP request multiplexer."""

    def __init__(self) -> None:
        self._handlers: dict[str, Handler] = {}

    def Handle(self, pattern: str, handler: Handler) -> None:
        """Handle registers the handler for the given pattern."""
        self._handlers[pattern] = handler

    def HandleFunc(self, pattern: str, handler: HandlerFunc) -> None:
        """HandleFunc registers the handler function for the given pattern."""
        self._handlers[pattern] = handler

    def ServeHTTP(self, w: ResponseWriter, r: Request) -> None:
        """ServeHTTP dispatches the request to the handler."""
        path = urllib.parse.urlparse(r.URL).path

        handler = None
        for pattern, h in self._handlers.items():
            if path == pattern or path.startswith(pattern.rstrip("/") + "/"):
                handler = h
                break

        if handler is None:
            NotFound(w, r)
            return

        handler(w, r)


class Server:
    """HTTP server."""

    def __init__(self, addr: str = ":8080", handler: ServeMux | None = None):
        self.Addr = addr
        self.Handler = handler or ServeMux()
        self._server: _http_server.HTTPServer | None = None

    def ListenAndServe(self) -> Result[None, GoError]:
        """ListenAndServe listens and serves HTTP requests."""
        try:
            host, port = self._parse_addr()
            mux = self.Handler

            class _MuxHTTPHandler(_http_server.BaseHTTPRequestHandler):
                def do_GET(self) -> None:
                    self._handle_request()

                def do_POST(self) -> None:
                    self._handle_request()

                def do_PUT(self) -> None:
                    self._handle_request()

                def do_DELETE(self) -> None:
                    self._handle_request()

                def do_HEAD(self) -> None:
                    self._handle_request()

                def do_OPTIONS(self) -> None:
                    self._handle_request()

                def do_PATCH(self) -> None:
                    self._handle_request()

                def _handle_request(self) -> None:
                    content_length = int(self.headers.get("Content-Length", 0))
                    body = self.rfile.read(content_length) if content_length > 0 else None

                    req = Request(
                        Method=self.command,
                        URL=self.path,
                        Header=Header(self.headers.items()),
                        Body=body,
                        Host=self.headers.get("Host", ""),
                    )

                    w = ResponseWriter(self)
                    try:
                        mux.ServeHTTP(w, req)
                    except Exception as e:
                        if not w._headers_written:
                            self.send_error(500, str(e))

                def log_message(self, format: str, *args: Any) -> None:
                    pass

            self._server = _http_server.HTTPServer((host, port), _MuxHTTPHandler)
            self._server.serve_forever()
            return Ok(None)
        except Exception as e:
            return Err(GoError(str(e), "http.Error"))

    def Shutdown(self) -> Result[None, GoError]:
        """Shutdown gracefully shuts down the server."""
        if self._server:
            self._server.shutdown()
        return Ok(None)

    def _parse_addr(self) -> tuple[str, int]:
        if ":" in self.Addr:
            parts = self.Addr.rsplit(":", 1)
            host = parts[0] if parts[0] else ""
            port = int(parts[1])
        else:
            host = ""
            port = int(self.Addr)
        return host, port


def HandleFunc(pattern: str, handler: HandlerFunc) -> None:
    """HandleFunc registers the handler function for the given pattern."""
    _handlers[pattern] = handler


def Handle(pattern: str, handler: Handler) -> None:
    """Handle registers the handler for the given pattern."""
    _handlers[pattern] = handler


def ListenAndServe(addr: str, handler: ServeMux | None = None) -> Result[None, GoError]:
    """ListenAndServe listens and serves HTTP requests."""
    server = Server(addr, handler)
    return server.ListenAndServe()


def FileServer(root: str) -> Handler:
    """FileServer returns a handler that serves HTTP requests with files from root."""
    import mimetypes
    import os

    def handler(w: ResponseWriter, r: Request) -> None:
        path = urllib.parse.urlparse(r.URL).path.lstrip("/")
        filepath = os.path.join(root, path)

        if os.path.isdir(filepath):
            filepath = os.path.join(filepath, "index.html")

        if not os.path.exists(filepath):
            NotFound(w, r)
            return

        content_type, _ = mimetypes.guess_type(filepath)
        if content_type:
            w.Header().Set("Content-Type", content_type)

        with open(filepath, "rb") as f:
            w.Write(f.read())

    return handler


def NotFound(w: ResponseWriter, r: Request) -> None:
    """NotFound replies with an HTTP 404 not found error."""
    Error(w, "404 page not found", StatusNotFound)


def Redirect(w: ResponseWriter, r: Request, url: str, code: int) -> None:
    """Redirect replies with a redirect to url."""
    w.Header().Set("Location", url)
    w.WriteHeader(code)


def Error(w: ResponseWriter, error: str, code: int) -> None:
    """Error replies with the specified error message and HTTP code."""
    w.Header().Set("Content-Type", "text/plain; charset=utf-8")
    w.Header().Set("X-Content-Type-Options", "nosniff")
    w.WriteHeader(code)
    w.Write(error.encode())



"""Go utf8 package bindings.

This module provides Python bindings for Go's unicode/utf8 package.
"""

from __future__ import annotations

import ctypes

from goated._core import get_lib, is_library_available

__all__ = [
    "DecodeLastRune",
    "DecodeLastRuneInString",
    "DecodeRune",
    "DecodeRuneInString",
    "EncodeRune",
    "FullRune",
    "FullRuneInString",
    "RuneCount",
    "RuneCountInString",
    "RuneLen",
    "RuneStart",
    "Valid",
    "ValidRune",
    "ValidString",
    # Constants
    "RuneError",
    "RuneSelf",
    "MaxRune",
    "UTFMax",
]

# Constants
RuneError = "\ufffd"  # The "error" rune or "Unicode replacement character"
RuneSelf = 0x80  # Characters below RuneSelf are represented as themselves
MaxRune = 0x10FFFF  # Maximum valid Unicode code point
UTFMax = 4  # Maximum number of bytes of a UTF-8 encoded character


def _encode(s: str) -> bytes:
    return s.encode("utf-8")


def _decode(b: bytes | None) -> str:
    if b is None:
        return ""
    return b.decode("utf-8")


def _get_rune(r: str | int) -> int:
    """Convert a string character to its code point (rune)."""
    if isinstance(r, int):
        return r
    if isinstance(r, str) and len(r) > 0:
        return ord(r[0])
    return 0


_fn_configured: set[str] = set()


def _configure_fn(lib: object, name: str, argtypes: list[type], restype: type) -> None:
    if name in _fn_configured:
        return
    fn = getattr(lib, name)
    fn.argtypes = argtypes
    fn.restype = restype
    _fn_configured.add(name)


def DecodeLastRune(p: bytes) -> tuple[str, int]:
    """DecodeLastRune unpacks the last UTF-8 encoding in p."""
    if not p:
        return RuneError, 0

    i = len(p) - 1
    while i > 0 and (p[i] & 0xC0) == 0x80:
        i -= 1
        if len(p) - i > UTFMax:
            return RuneError, 1

    try:
        s = p[i:].decode("utf-8")
        if s:
            return s[0], len(p) - i
    except UnicodeDecodeError:
        pass

    return RuneError, 1


def DecodeLastRuneInString(s: str) -> tuple[str, int]:
    """DecodeLastRuneInString is like DecodeLastRune but for strings."""
    if not s:
        return RuneError, 0

    last_char = s[-1]
    byte_len = len(last_char.encode("utf-8"))
    return last_char, byte_len


def DecodeRune(p: bytes) -> tuple[str, int]:
    """DecodeRune unpacks the first UTF-8 encoding in p."""
    if not p:
        return RuneError, 0

    for width in range(1, min(UTFMax + 1, len(p) + 1)):
        try:
            s = p[:width].decode("utf-8")
            if s:
                return s[0], width
        except UnicodeDecodeError:
            continue

    return RuneError, 1


def DecodeRuneInString(s: str) -> tuple[str, int]:
    """DecodeRuneInString is like DecodeRune but for strings."""
    if not s:
        return RuneError, 0

    first_char = s[0]
    byte_len = len(first_char.encode("utf-8"))
    return first_char, byte_len


def EncodeRune(p: bytearray, r: str | int) -> int:
    """EncodeRune writes the UTF-8 encoding of the rune into p."""
    rune = _get_rune(r)

    if rune < 0 or rune > MaxRune or (0xD800 <= rune <= 0xDFFF):
        rune = ord(RuneError)

    try:
        encoded = chr(rune).encode("utf-8")
        for i, b in enumerate(encoded):
            if i < len(p):
                p[i] = b
        return len(encoded)
    except (ValueError, OverflowError):
        encoded = RuneError.encode("utf-8")
        for i, b in enumerate(encoded):
            if i < len(p):
                p[i] = b
        return len(encoded)


def FullRune(p: bytes) -> bool:
    """FullRune reports whether p begins with a full UTF-8 encoding of a rune."""
    if not p:
        return False

    first_byte = p[0]
    if first_byte < 0x80 or first_byte < 0xC0:
        return True
    elif first_byte < 0xE0:
        expected = 2
    elif first_byte < 0xF0:
        expected = 3
    elif first_byte < 0xF8:
        expected = 4
    else:
        return True

    return len(p) >= expected


def FullRuneInString(s: str) -> bool:
    """FullRuneInString is like FullRune but for strings."""
    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(lib, "goated_utf8_FullRuneInString", [ctypes.c_char_p], ctypes.c_bool)
            return bool(lib.goated_utf8_FullRuneInString(_encode(s)))
        except Exception:
            pass

    return len(s) > 0


def RuneCount(p: bytes) -> int:
    """RuneCount returns the number of runes in p."""
    try:
        return len(p.decode("utf-8", errors="replace"))
    except Exception:
        return len(p)


def RuneCountInString(s: str) -> int:
    """RuneCountInString returns the number of runes in s."""
    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(
                lib, "goated_utf8_RuneCountInString", [ctypes.c_char_p], ctypes.c_longlong
            )
            return int(lib.goated_utf8_RuneCountInString(_encode(s)))
        except Exception:
            pass

    return len(s)


def RuneLen(r: str | int) -> int:
    """RuneLen returns the number of bytes in the UTF-8 encoding of the rune."""
    rune = _get_rune(r)

    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(lib, "goated_utf8_RuneLen", [ctypes.c_int], ctypes.c_longlong)
            return int(lib.goated_utf8_RuneLen(rune))
        except Exception:
            pass

    # Pure Python fallback
    if rune < 0 or rune > MaxRune or (0xD800 <= rune <= 0xDFFF):
        return -1

    if rune < 0x80:
        return 1
    elif rune < 0x800:
        return 2
    elif rune < 0x10000:
        return 3
    else:
        return 4


def RuneStart(b: int) -> bool:
    """RuneStart reports whether b could be the first byte of an encoded rune."""
    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(lib, "goated_utf8_RuneStart", [ctypes.c_ubyte], ctypes.c_bool)
            return bool(lib.goated_utf8_RuneStart(b))
        except Exception:
            pass

    return (b & 0xC0) != 0x80


def Valid(p: bytes) -> bool:
    """Valid reports whether p consists entirely of valid UTF-8-encoded runes."""
    try:
        p.decode("utf-8")
        return True
    except UnicodeDecodeError:
        return False


def ValidRune(r: str | int) -> bool:
    """ValidRune reports whether r can be legally encoded as UTF-8."""
    rune = _get_rune(r)

    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(lib, "goated_utf8_ValidRune", [ctypes.c_int], ctypes.c_bool)
            return bool(lib.goated_utf8_ValidRune(rune))
        except Exception:
            pass

    if rune < 0 or rune > MaxRune:
        return False
    return not 55296 <= rune <= 57343


def ValidString(s: str) -> bool:
    """ValidString reports whether s consists entirely of valid UTF-8-encoded runes."""
    if is_library_available():
        try:
            lib = get_lib()
            _configure_fn(lib, "goated_utf8_ValidString", [ctypes.c_char_p], ctypes.c_bool)
            return bool(lib.goated_utf8_ValidString(_encode(s)))
        except Exception:
            pass

    try:
        s.encode("utf-8")
        return True
    except UnicodeEncodeError:
        return False



"""Drop-in replacement for Python's hashlib module, backed by Go's crypto packages.

Provides Go-accelerated one-shot hashing for md5, sha1, sha256, sha512.
Falls back to Python's hashlib for streaming updates and unsupported algorithms.

Usage:
    from goated.compat import hashlib
    digest = hashlib.sha256(b"data").hexdigest()
    # Or use fast one-shot:
    hex_str = hashlib.go_hexdigest("sha256", b"data")
"""

from __future__ import annotations

import ctypes
import hashlib as _hashlib
from collections.abc import Callable

# Re-export everything from Python's hashlib
from hashlib import (  # noqa: F401
    algorithms_available,
    algorithms_guaranteed,
    blake2b,
    blake2s,
    md5,
    new,
    sha1,
    sha3_224,
    sha3_256,
    sha3_384,
    sha3_512,
    sha224,
    sha256,
    sha384,
    sha512,
)
from typing import Any, BinaryIO

from goated._core import _USE_GO_LIB, get_lib

_lib_setup = False
_GO_FUNCS: dict[str, Any] = {}


def _setup_lib() -> None:
    global _lib_setup
    if _lib_setup or not _USE_GO_LIB:
        return
    try:
        lib = get_lib().lib
        for name in ("md5", "sha1", "sha256", "sha512"):
            fn = getattr(lib, f"goated_crypto_{name}_SumHex")
            fn.argtypes = [ctypes.c_char_p, ctypes.c_int64]
            fn.restype = ctypes.c_char_p
            _GO_FUNCS[name] = fn
        for name in ("sha1", "sha256", "sha512"):
            fn = getattr(lib, f"goated_hmac_{name}")
            fn.argtypes = [ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p, ctypes.c_int]
            fn.restype = ctypes.c_char_p
            _GO_FUNCS[f"hmac_{name}"] = fn
        _lib_setup = True
    except Exception:
        pass


def go_hexdigest(algorithm: str, data: bytes) -> str:
    """One-shot Go-accelerated hex digest. Faster than streaming for complete data."""
    _setup_lib()
    fn = _GO_FUNCS.get(algorithm)
    if fn is not None:
        result = fn(data, len(data))
        if result:
            return result.decode("ascii")
    return _hashlib.new(algorithm, data).hexdigest()


def go_hmac_hexdigest(algorithm: str, key: bytes, msg: bytes) -> str:
    """One-shot Go-accelerated HMAC hex digest."""
    _setup_lib()
    fn = _GO_FUNCS.get(f"hmac_{algorithm}")
    if fn is not None:
        result = fn(key, len(key), msg, len(msg))
        if result:
            return result.decode("ascii")
    import hmac as _hmac

    return _hmac.new(key, msg, algorithm).hexdigest()


def file_digest(
    fileobj: BinaryIO,
    digest: str | Callable[[], Any],
    *,
    _bufsize: int = 262144,
) -> Any:
    """Hash the contents of a file-like object."""
    h = new(digest) if isinstance(digest, str) else digest()
    buf = fileobj.read(_bufsize)
    while buf:
        h.update(buf)
        buf = fileobj.read(_bufsize)
    return h


__all__ = [
    "md5",
    "sha1",
    "sha224",
    "sha256",
    "sha384",
    "sha512",
    "sha3_224",
    "sha3_256",
    "sha3_384",
    "sha3_512",
    "blake2b",
    "blake2s",
    "new",
    "file_digest",
    "algorithms_available",
    "algorithms_guaranteed",
    "go_hexdigest",
    "go_hmac_hexdigest",
]
